"""Interface to source(s) of individual item response data,
e.g., a flat database or Excel sheet(s) or a csv file,
accessible by some pandas.read_xxx function, or similar user-defined function.

Item Responses are accessed via an iterator,
yielding a list of integer-encoded ordinal responses for each table row.
Each table row must include response data from ONE unique respondent.

*** Main Access Function
item_response_table --- creates interface to ONE source of item response data
default_file_reader --- selects pandas reader depending on file suffix

*** Classes

ItemResponseFile --- interface to single file containing item response data.
ItemResponseSQL --- interface to SQL database accessed by pd.read_sql using SQLAlchemy

ItemResponseTable --- superclass interface to all concrete sources of tabulated data
    acting as an iterator over data from individual respondents.

Tables --- iterable chaining several sub-iterables with tabulated data. *****

*** Version history:
* Version 0.6.0:
2023-07-xx, allow reading SQLAlchemy or sqlite3 database
2022-09-15, using pandas read_xxx functions for input data sources
"""
# **** test reading by SQLAlchemy.engine
# **** allow reading by chunks, as complete pd.DataFrame ? sqlite chunks?

import numpy as np
import pandas as pd
from pathlib import Path
from itertools import chain

import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)


# ------------------------------------------------ Reader exceptions
class ReadError(RuntimeError):
    """Any type of exception while attempting to read data from a source
    """


# ---------------------------------------------------------------
def item_response_table(items,
                        source,
                        sql=None,
                        accept_fcn=None,
                        sample_factor=1,
                        read_fcn=None,
                        **kwargs):
    """Factory creating an access object to a source of item response data.
        :param items: dict with (column_name, list(ordinal categories))
            for items and allowed item responses, with one column for each item.
        :param source: source for reading input data,
            e.g., a file Path, path string, or URL,
            OR an SQLAlchemy engine object, if reading from a general SQL database,
            OR a sqlite3.connection object, if using sqlite3 database
        :param sql: (optional) string with SQL SELECT statement, for pd.read_sql
            OR an SQL table name (if source is an SQLAlchemy engine)
            OR a SQLAlchemy Select object, generated by its select(...) function
        :param read_fcn: (optional) any user-defined file-reader function
            with signature similar to pd.read_excel or pd.read_csv.
            If None, a default pandas function is determined by file_path.suffix.
        :param accept_fcn: (optional) test function for each row to be valid input
        :param sample_factor: (optional) scalar integer for down-sampling among records
            Only one record is returned for every sample_factor records in the file
        :param kwargs: (optional) additional arguments for the reader function,
            depending on the selected reader function, e.g., for read_excel:
            index_col = column name or index for respondent ID,
                used to eliminate duplicate records from same respondent
            usecols = list of column names, if there are many redundant items
            dtype = dict with column datatype(s),
            converters = dict with elements (item_column, recoding_fcn)
            sheet_name = sheet name or list of sheet names, in case excel-type file
        :return: generator yielding item response arrays r, with
            r[i] = Response index for i-th Item, with origin=1,
            coded as integer in range 1,..., L_i for valid response, and
            r[i] = 0 for a missing response.
    """
    if sql is None:
        return ItemResponseFile(path=Path(source),
                                items=items, accept_fcn=accept_fcn, sample_factor=sample_factor,
                                read_fcn=read_fcn, **kwargs)
    else:  # con must be SQLAlchemy engine or sqlite3.connection object
        return ItemResponseSQL(con=source, sql=sql,
                               items=items, accept_fcn=accept_fcn, sample_factor=sample_factor,
                               **kwargs)


def default_file_reader(file_path):
    """Find a function that can read the given file path
    :param file_path: a Path instance
    :return: a pandas reader function, e.g., pd.read_excel
    """
    suffix = file_path.suffix
    if suffix in ['.xlsx', '.xls', '.odf', '.ods', '.odt']:
        return pd.read_excel
    elif suffix in ['.csv', '.txt']:
        return pd.read_csv
    elif suffix in ['.sav', '.zsav']:
        return pd.read_spss
    elif suffix in ['.dta']:
        return pd.read_stata
    else:
        raise ReadError(f'Must assign explicit read_fcn to read file format {suffix}')


# ---------------------------------------------- interface classes
class ItemResponseTable:
    """Superclass interface to a source of ordinal item responses,
    that can be accessed by some pandas.read_xxx or similar functions.
    Item Response data must be stored in rows and named items,
    e.g., a csv or Excel file.
    """
    def __init__(self, items, accept_fcn=None, sample_factor=1):
        """
        :param items: dict with (column_name, list(ordinal categories))
            for item responses to be returned to caller
        :param accept_fcn: (optional) test function for each row to be valid input
        :param sample_factor: (optional) scalar integer for down-sampling among records
            Only one record is returned for every sample_factor records in the file
        :return: generator yielding item response arrays r, with
            r[i] = Response index for i-th Item, with origin=1,
            coded as integer in range 1,..., L_i for valid response, and
            r[i] = 0 for a missing response.
        """
        self.items = items
        if accept_fcn is None:
            accept_fcn = _default_accept_all  # lambda row: True
        self.accept_fcn = accept_fcn
        self.sample_factor = sample_factor

    def __repr__(self):
        return (self.__class__.__name__ + '(\n\t'
                + ',\n\t'.join((f'{k}={repr(v)}' for (k, v) in self.__dict__.items()))
                + ')')

    @property
    def column_dtypes(self):
        return {c: pd.CategoricalDtype(categories=list(cats),
                                       ordered=True)
                for (c, cats) in self.items.items()}

    def __len__(self):
        """Number of (down-sampled) data records in file,
        not including header,
        exact or crudely estimated.
        This default method simply reads and counts all rows.
        Subclasses may implement faster method.
        :return: scalar integer
        """
        df = self.load()
        try:
            if type(df) is pd.DataFrame:
                return df.shape[0] // self.sample_factor
            elif type(df) is dict:  # dict of sheets from pd.read_excel()
                return sum(df_i.shape[0]
                           for df_i in df.values()) // self.sample_factor
            else:  # iterable of DataFrame objects, if reading by chunks
                return sum(df_i.shape[0]
                           for df_i in df) // self.sample_factor
        except RuntimeError as e:
            raise ReadError(f'Cannot count input data; Error: {e}')

    def __iter__(self):
        """Iterator over all ordinal item-response vectors from source defined by self.
        :return: generator yielding item response vectors r, with
            r[i] = Ordinal Response Code for i-th Item,
            coded as integer in range 0,..., L_i-1 for valid response, and
            r[i] = -1 for a missing response.
        """
        def gen_codes(df):
            """Generate code vectors from rows a DataFrame instance
            :param df: a pd.DataFrame instance
            :return: iterator
            """
            df = _check_unique_index(df)
            try:
                df_codes = [df[c].astype(dt).array.codes
                            for (c, dt) in self.column_dtypes.items()]
                df_codes = np.array(df_codes).T
                return (r for (i, r) in enumerate(df_codes)
                        if self.accept_fcn(r) and 0 == i % self.sample_factor)
            except RuntimeError as e:
                raise ReadError(f'Cannot extract code vectors from {df}; Error: {e}')
        # ---------------------------------------------------------

        df = self.load()
        if type(df) is pd.DataFrame:
            return gen_codes(df)
        elif type(df) is dict:  # dict of sheets from pd.read_excel()
            return chain.from_iterable((gen_codes(df_s)
                                        for (s, df_s) in df.items()))
        else:  # iterable of DataFrame objects, if reading by chunks
            try:
                return chain.from_iterable((gen_codes(df_i) for df_i in df))
            except RuntimeError as e:
                raise ReadError(f'Cannot use input data; Error: {e}')

    def load(self):
        """Abstract method to load all data from a file or other source
        :return: a pd.DataFrame or dict of pd.DataFrame objects
        """
        raise NotImplementedError


# -----------------------------------------------------------
class ItemResponseFile(ItemResponseTable):
    """Access to file storage of item response data,
    e.g., xlsx, csv, or similar
    """
    def __init__(self, items, path, accept_fcn=None, sample_factor=1, read_fcn=None, **file_kwargs):
        """
        :param items: dict with (column_name, list(ordinal categories))
            for item responses to be returned to caller
        :param path: Path instance defining input file
        :param accept_fcn: (optional) test function for each row to be valid input
        :param sample_factor: (optional) scalar integer for down-sampling among records
            Only one record is returned for every sample_factor records in the file
        :param read_fcn: (optional) reader function, e.g., pd.read_excel,
            or any user-supplied function with similar signature.
            If None, a default pandas function is determined by file_path.suffix.
        :param file_kwargs: (optional) additional arguments for the reader function, e.g.,
            usecols = list of column names, if there are many redundant items
            dtype = dict with column datatype(s),
            converters = dict with elements (item_colum, recoding_fcn)
            sheet_name = single sheet name to access, in case excel-type file
        """
        super().__init__(items=items, accept_fcn=accept_fcn, sample_factor=sample_factor)
        self.path = path
        if read_fcn is None:
            read_fcn = default_file_reader(path)
        self.read_fcn = read_fcn
        self.file_kwargs = file_kwargs

    def load(self):
        """Load data from file into local memory
        :return: a pd.DataFrame object or dict with such objects (Excel file sheets).
        """
        logger.debug(f'Loading from {self.path} with {self.read_fcn}')
        try:
            return self.read_fcn(self.path,
                                 **self.file_kwargs)
        except Exception as e:
            raise ReadError(f'{self.read_fcn.__name__}({self.path}) error: {e}')


# -------------------------------------------------------
class ItemResponseSQL(ItemResponseTable):
    """Access class using SQLAlchemy specification of database storage.
    """
    def __init__(self, con, sql, items, accept_fcn=None, sample_factor=1, **kwargs):
        """
        :param con: SQLAlchemy engine instance (preferably),
            OR SQLAlchemy or sqlite3 connection instance, input for pd.read_sql
        :param sql: SQL select string or SQLAlchemy selectable, input for pd.read_sql
        :param items: dict with (column_name, list(ordinal categories))
            for item responses to be returned to caller
        :param accept_fcn: (optional) test function for each row to be valid input
        :param sample_factor: (optional) scalar integer for down-sampling among records
            Only one record is returned for every sample_factor records in the file
        :param kwargs: (optional) additional arguments for pd.read_sql, e.g.,
            columns = list of column names, if there are many redundant items
            index_col = column for respondent ID
            chunksize = number of rows for each input chunk
            params = any other database-dependent parameters
        """
        # ******* should CLOSE con instance if it is sqlite3.connection instance
        super().__init__(items=items, accept_fcn=accept_fcn, sample_factor=sample_factor)
        self.con = con
        self.sql = sql
        self.kwargs = kwargs

    def load(self):
        """Load data using pd.read_sql with SQLAlchemy instances as input
         :return: pd.DataFrame or iterator of DataFrame objects
         """
        logger.debug(f'Loading read_sql({self.sql}, {self.con}')
        try:
            return pd.read_sql(sql=self.sql, con=self.con,
                               **self.kwargs)
        except Exception as e:
            raise ReadError(f'read_sql({self.sql}, {self.con}) error: {e}')


class Tables:
    """Iterable chaining sub-iterables
    """
    def __init__(self, *table_list):
        """
        :param table_list: list of iterables, e.g., lists or ItemResponseTable objects.
        """
        self.table_list = table_list

    def __len__(self):
        """Total number of data records in all included sub-tables,
        exact or crudely estimated.
        :return: scalar integer
        """
        return sum(len(t) for t in self.table_list)

    def __iter__(self):
        """iterator of records in all included sub-tables
        :return: iterator over all response records in all sub-tables
        """
        return chain.from_iterable(self.table_list)


# ----------------------------------------------------- Module help functions:
def _default_accept_all(r):
    """Default function for ItemResponseTable property accept_fcn.
    :param r: list with one record from input
    :return: always True, without any checking
    """
    return True


def _check_unique_index(df):
    """Check if input dataframe has unique index labels,
    and remove duplicates
    :param df: a pd.DataFrame instance
    :return: df_unique = corresponding pd.DataFrame with duplicates removed
    """
    if df.index.is_unique:
        return df
    else:
        n_unique = len(df.index.unique())
        n_df = df.shape[0]
        logger.debug(f'DataFrame has {n_unique} unique index labels, less than all {n_df} rows. ' +
                       'Keeping last record from each respondent.')
        df_u = df.groupby(level=[0]).last()
        return df_u


# -------------------------------------------------------- Module TEST:
if __name__ == '__main__':
    # from pathlib import Path
    import sqlite3
    import ir_logging


    HAQ_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry/IRT-IOI-HA/IOI-HA-Data'
    # ioiha_data_path = HAQ_PATH / 'IRT-IOI-HA' / 'IOI-HA-data'
    # to ioi-ha data sets OTHER THAN the Swedish Quality Registry
    # ------------------------------------------------------------------

    print('\n*** Testing item_response_table with Hickson data set:\n')
    work_path = HAQ_PATH
    test_file = HAQ_PATH / 'Hickson' / 'Short_Eartrak AUS.xlsx'

    # ------------------------------------- setup logging for test:
    ir_logging.setup(work_path / 'test_ir_source.txt')
    logger = logging.getLogger(__name__)

    column_names = [f'Q0{i}' for i in range(1, 8)]
    ir_hickson = item_response_table(items={c: [1, 2, 3, 4, 5]
                                          for c in column_names},
                                   source=test_file, index_col='EarRowId')  #0)
    print('ir_hickson = ', ir_hickson)

    ir_df = ir_hickson.load()
    print('ir_df.head =\n', ir_df.head(10))

    print('printing a few rows:')
    n = 0
    for r in ir_hickson:
        print(r)
        n += 1
        if n > 10:
            break

    print(f'len(ir_hickson) = ', len(ir_hickson))

    sql_path = work_path / 'test_sql'
    sql_path.mkdir(parents=True, exist_ok=True)
    # *** create sqlite3 database with copy of ir_df:
    # con = sqlite3.connect(sql_path / 'testHickson.db' )
    # cur = con.cursor()
    # # cur.execute("CREATE TABLE ioiha(" + ','.join(ir_df.columns) + ")")
    # res = cur.execute("SELECT name FROM sqlite_master")
    # print(res.fetchone())

    # cur.execute("""
    # INSERT INTO ioiha VALUES
    #     ('tes1', 'arne', 1, 2, 3, 4, 5, 6 ,7),
    #     ('test2', 'eddie', 1, 7.5, 3, 4, 5, 6, 9) """)
    # con.commit()
    # res = cur.execute("SELECT * FROM ioiha")
    # r = res.fetchall()
    # print(r)
    # con.close()

    # print('*** test read_sql')
    # con = sqlite3.connect(sql_path / 'testHickson.db' )
    # ir_test = pd.read_sql_query("SELECT * FROM ioiha", con, index_col='EarRowId')
    # print(ir_test)
    # # con.close()

    print('*** test Hickson to_sql')

    con = sqlite3.connect(sql_path / 'testHickson.db' )
    n_rows = ir_df.to_sql('ioiha', con, index=True, if_exists='replace')
    print(f'{n_rows} saved by Pandas to_sql')
    # NOTE: Hickson's xlsx has integer codes and '.' for missing data, so the result is mixed: dtype= object
    # When saving to sqlite by to_sql, dtype object -> string in all items, except Q00 which is still integer
    # SQL requires all elements in one column to have SAME dtype

    # print('*** test read_sql_query')
    # ir_test = pd.read_sql_query("SELECT * FROM ioiha", con, index_col='EarRowId')
    # print(ir_test)
    # con.close()

    print('*** test read_sql')
    con = sqlite3.connect(sql_path / 'testHickson.db' )
    ir_test = pd.read_sql("SELECT * FROM ioiha", con, index_col='EarRowId')
    print(ir_test)
    con.close()

    print('\n*** Test item_response_table using sqlite3 connection and query input ***')
    con = sqlite3.connect(sql_path / 'testHickson.db' )
    ir_sql = item_response_table(source=con,
                                   sql="SELECT * FROM ioiha",
                                   index_col='EarRowId',
                                   items={c: [str(i) for i in range(1, 6)]
                                          for c in column_names}
                                   )  # NOTE: string-encoded integers here
    ir_df = ir_sql.load()
    print(ir_df.head(20))

    print('printing a few rows:')
    n = 0
    for r in ir_sql:
        print(r)
        n += 1
        if n > 10:
            break

    con.close()

    print('\n*** Test SQLAlchemy input directly ***')
    from sqlalchemy import create_engine
    engine = create_engine("sqlite+pysqlite:///" + str(sql_path / 'testHickson.db'))
    from sqlalchemy import text
    with engine.connect() as conn:
        result = conn.execute(text("select * from ioiha")).fetchmany(20)
        print(np.array(result))

    print('\n*** Test item_response_table using SQLAlchemy engine ***')
    ir_sql = item_response_table(source=engine,
                                   sql='ioiha',  # *** "SELECT * FROM ioiha",
                                   index_col='EarRowId',
                                   items={c: [str(i) for i in range(1, 6)]
                                          for c in column_names}
                                   )  # NOTE: string-encoded integers here
    ir_df = ir_sql.load()
    print(ir_df.head(20))

    print('finished')





"""General utility classes and functions for internal use.

Defines class
ItemResponseBase --- container for common properties
    specifying the indexing of separate types of model parameters, and
    calculating logprob for observed data, given tentative model parameters.


*** Version history:
* Version 1.0.0: New module for this version. Modified after EmaCalc.ema_base.
"""
import numpy as np
# from itertools import chain
# from collections import namedtuple
# from scipy.special import logit, expit
# from scipy.special import logsumexp, softmax
import logging
# import pandas as pd

from .ir_thresholds import ThresholdsFree, ThresholdsMidFixed


# ------------------------------------------------------
__version__ = "2023-07-20"

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST

PRIOR_PSEUDO_RESPONDENT = 0.5  # seems to work OK
# = hyperprior total pseudo-count re ONE real respondent
# = prior GaussianRV.mean.learned_weight.
# This value is NOT critical for Occam's Razor behavior.

PRIOR_PARAM_SCALE = 1.
# = main hyperprior scale of most Gaussian model parameters,
# defined in LatentNormal d-prime (probit) units for attribute parameters,
# rescaled if the LatentLogistic (logistic) model is used for latent variables.

PRIOR_PREC_A = PRIOR_PSEUDO_RESPONDENT / 2
# = GaussianRV precision shape for ALL parameters

PRIOR_PREC_B = PRIOR_PARAM_SCALE**2 / 2
# = GaussianRV precision inv_scale for MOST model parameters
# -> allows very small precision <=> large inter-individual variance
# -> mode of global_pop component-element variance = PRIOR_PREC_B /(PRIOR_PREC_A + 1) approx= PRIOR_PREC_B


# ------------------------------------------------------- Help classes
class Slicer:
    """Help class to create consecutive slices for parts of parameter vector
    """
    def __init__(self, start=0):
        self.start = start

    def slice(self, length):
        """Create a new slice object tight after previous slice
        :param length: desired slice size
        :return: a slice object
        """
        stop = self.start + length
        sl = slice(self.start, stop)
        self.start = stop  # for next call
        return sl


# ------------------------------------------------------------------
class ItemResponseBase:
    """Container for common properties and methods for OrdinalItemResponseModel and its parts.

    Each individual parameter distribution is represented by
    a large set of samples, stored as property ir_group.ResponseGroup.xi, with
    xi[n, :] = n-th sample of (eta, theta) parameters for ONE GROUP of respondents.

    All model classes share mapping properties defined here,
    for extracting parameter subsets for item thresholds and individual traits
    from a xi array of parameter vectors.
    Parameter types are stored consecutively in order (eta, theta)
    """
    def __init__(self, rm_class, thr, eta_slices):  #, restrict_threshold):
        """
        :param rm_class: response model class, e.g., ir_graded_response_prob.GradedResponse,
            calculating log probability of any response given model parameters
        :param thr: class defining threshold calculations, defined in module ir_thresholds,
        :param eta_slices: list of slice objects, such that
            eta_i = xi[:, eta_slices[i]] = array of thresholds for i-th item, by
            tau_i = self.thr.tau(eta_i)
        """
        self.rm_class = rm_class
        self.thr = thr
        self.eta_slices = eta_slices
        # self.restrict_threshold = restrict_threshold  # not needed ?

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    @classmethod
    def initialize(cls, ir_count, restrict_threshold, rm_class):
        """Assign all parameter-extraction slices, and
        :param ir_count: list of 1D arrays with response counts
            ir_count[i][l] = number of l-th ordinal responses to i-th item
            *** need only sizes, not actual counts ***
        :param restrict_threshold: boolean switch
            to force restriction on response-threshold locations
        :param rm_class: response model class, e.g., ir_graded_response_prob.GradedResponse,
            calculating log probability of any response given model parameters
        :return: a cls instance

        Result: all properties initialized
        NOTE: xi parameter slices in order: all alpha, all beta, all eta
        """
        if restrict_threshold:
            thr = ThresholdsMidFixed
        else:
            thr = ThresholdsFree
        # Number of eta parameters depends on class thr

        slice_gen = Slicer()  # creating adjacent slice objects
        eta_slices = [slice_gen.slice(thr.n_param(len(count_i)))
                      for count_i in ir_count]
        # = list of slice objects, such that
        # eta_i = xi[:, eta_slices[i]] = array defining thresholds for i-th item, by
        # tau_i = self.thr.tau(eta_i)

        # **** allow tied thresholds for several items, like EmaCalc ??? ************
        # tied_scales = None
        # if len(attribute_slices) > len(unique_slices):
        #     attribute_keys = list(emf.attribute_scales.keys())
        #     tied_scales = [[attribute_keys.index(a_key) for a_key in a_keys]
        #                    for a_keys in emf.scale_attributes.values()]
        # tied_scales[s] = list of attribute indices a_i, such that
        # attribute_slices[a_i].eta_slice is eta_slices[s],
        # allowing same eta parameters to be used for more than one attribute

        return cls(rm_class,
                   thr,
                   eta_slices)

    @property
    def eta_all_slice(self):
        return slice(self.eta_slices[0].start, self.eta_slices[-1].stop)

    @property
    def theta_slice(self):
        """All remaining elements are theta
        """
        return slice(self.eta_slices[-1].stop, None)

    def n_eta(self):
        """Total number of eta parameters in model
        :return: scalar
        """
        return self.eta_slices[-1].stop

    def eta_all(self, xi):
        """Extract all threshold-defining parameters eta from given sample array
        :param xi: 2D array of (eta, theta) samples
        :return: 2D array of eta samples
        """
        return xi[:, self.eta_all_slice]

    def theta(self, xi, n_subjects):
        """Extract all theta parameters from given sample array
        :param xi: 2D array of (eta, theta) samples
        :param n_subjects: scalar number of respondents included in xi[:, self.theta_slice
        :return: 3D array of theta samples
            theta[n, s, d] = n-th sample of d-th trait for s-th respondent
        """
        n_samples = xi.shape[0]
        return xi[:, self.theta_slice].reshape((n_samples, n_subjects, -1))

    def tau(self, xi):
        """Extract threshold vectors for all items
        :param xi: 1D or 2D array of (eta, theta) samples
        :return: list of threshold samples tau_i, INCL extreme -inf, +inf
            tau_i[n, l] = n-th sample of l-th threshold for
        """
        return [self.thr.tau(xi[..., i])
                for i in self.eta_slices]

    # ---------------------- methods for logprob calculation:
    def item_response_logprob(self, r, eta, theta):
        """Rating logprob, given parameter values
        :param r: array of integer-coded responses
            r[i, s] = ordinal response to i-th item by s-th respondent
        :param eta: 2D array of tentative threshold-defining parameters, same for all respondents
            eta[n, self.eta_slice[i]] = n-th sample of eta for i-th item
        :param theta: 3D array of tentative trait values
            theta[n, s, d] = n-th sample of d-th trait for s-th respondent
            eta.shape[0] == theta.shape[0] == n_samples, joint for (eta, theta)
            theta.shape[1] == r.shape[1] == number of respondents
        :return: lp = 3D array, with elements
            lp[i, n, d] = sum_s log P(r[i, s] | eta_i[n, :], theta[n, s, d]
        """
        return np.array([self.rm_class.item_sum_response_logprob(r_i,
                                                                 self.thr.tau(eta[:, i]),
                                                                 theta)
                         for (i, r_i) in zip(self.eta_slices, r)])

    def response_logprob(self, r, eta, theta, weight):
        """Rating logprob given tentative parameter values, weighted by item-by-trait prob.
        Help method for group model learning.
        :param r: array of integer-coded responses
            r[i, s] = ordinal response to i-th item by s-th respondent
        :param eta: 2D array of tentative threshold-defining parameters for all items,
            eta[n, self.eta_slice[i]] = n-th sample of eta for i-th item
        :param theta: 3D array of tentative trait values
            theta[n, s, d] = n-th sample of d-th trait for s-th respondent
            eta.shape[0] == theta.shape[0] == n_samples (jointly sampled)
            theta.shape[1] == responses.shape[1]
        :param weight: array with item-trait mapping
            weight[i, d] = prob that response to i-th item is determined by d-th trait
        :return: ll = 1D array
            ll[n] = sum_{i, s} log P{r[i, s] | xi[n, :]}
            ll.shape == xi.shape[:-1]
        """
        return np.sum(self.item_response_logprob(r, eta, theta) * weight[:, None, :],
                      axis=(0, 2))

    def d_response_logprob(self, r, eta, theta, weight):
        """Gradients of rating_logprob(...)
        :param eta: 2D array of tentative threshold-defining parameters for all items,
            eta[n, self.eta_slice[i]] = n-th sample of eta for i-th item
        :param theta: 3D array of tentative trait values
            theta[n, s, d] = n-th sample of d-th trait for s-th respondent
            eta.shape[0] == theta.shape[0] == n_samples (joint)
        :param r: array of integer-coded responses
            r[i, s] = ordinal response to i-th item by s-th respondent
            theta.shape[1] == responses.shape[1]
        :param weight: array with item-trait mapping
            weight[i, d] = prob that response to i-th item determined by d-th trait
        :return: d_ll = 2D array with elements
            d_ll[n, k] = d response_logprob(r, eta, theta, weight)[n] / d xi[n, k]
            d_ll.shape == self.xi.shape
            such that
            d_ll[n, k] = d response_logprob(r, eta, theta, weight)[n] / d eta[n, k]; k in self.eta_slice_all
            d_ll[n, k] = d response_logprob(r, eta, theta, weight)[n] / d theta[n, sk]; sk in self.theta_slice_all
        """
        dll = [self.rm_class.d_item_sum_response_logprob(r_i, self.thr.tau(eta[:, i]), theta)
               for (i, r_i) in zip(self.eta_slices, r)]
        # dll[i][0][n, k, d] = d item_response_logprob(r_i, tau_i[n, k], theta[n,..., d)[n, d] / d tau_i[n, k]
        # dll[i][1][n, s, d] = d item_response_logprob(r_i, ...)[n, d] / d theta[n, s, d] )
        dll_tau = [np.dot(dll_i[0], w_i)
                   for (dll_i, w_i) in zip(dll, weight)]
        # dll_tau[i][n, l] = d response_logprob(r[i] | tau_i[n], theta[n]) / d tau_i[n, l]
        # dll_eta_test = [np.einsum('nl, nlj -> nj',
        #                           dll_i, self.thr.d_tau(eta[:, i]))
        #                 for (dll_i, i) in zip(dll_tau, self.eta_slices)]
        dll_eta = [np.sum(dll_i[..., None] * self.thr.d_tau(eta[:, i]), axis=1)
                   for (dll_i, i) in zip(dll_tau, self.eta_slices)]
        # dll_eta[i][n, k] = d response_logprob(r[i] | tau_i[n], theta[n]) / d eta[n, k]
        dll_theta = sum(dll_i[1] * w_i
                        for (dll_i, w_i) in zip(dll, weight))  # sum across items
        # dll_theta[n, s, d] = sum_i d item_response_logprob(r_i, tau_i, theta[n, s, d) / d theta[n, s, d] )
        n_samples = dll_theta.shape[0]
        return np.concatenate(dll_eta + [dll_theta.reshape((n_samples, -1))],
                              axis=-1)


# ------------------------------------------------- TEST:



"""This module implements a random matrix with Wishart distribution
suitable to represent the precision matrix of a Gaussian random vector
with correlated elements.

This version is used as part of ItemResponseCalc package.

*** Main Classes:

WishartRV -- a random symmetric matrix with Wishart distribution.

Similar to a frozen scipy.stats.wishart distribution object,
but extends functionality by allowing the objects to adapt to observed data.

*** Version History:
* Version 1.0.0: New module in this version of ItemResponseCalc
"""
import numpy as np
from scipy.special import multigammaln, psi
from scipy.stats import wishart


# ---------------------------------------------------------------------------
class WishartRV:
    """A symmetric precision matrix (called Lambda in doc) modelled as a random variable,
    with a probability density function
    p(Lambda) = (1/C) | Lambda|**((nu - p - 1)/2) exp(- trace(inv(V) Lambda)/2),
    where
        V = symmetric positive definite scale matrix,
        p = length of associated Gaussian vector
        nu = degree of freedom parameter > p - 1,
        Lambda.shape == V.shape == (p, p)
        C = |V|**(nu/2) 2**(nu * p / 2) Gamma_p(nu/2) = the normalization factor
        Gamma_p = multivariate gamma function
    The distribution is proper for nu > p - 1.
    """
    def __init__(self, scale, df):
        """
        :param scale: 2D array or array-like symmetric scale matrix, called V in doc
        :param df: scalar degrees of freedom, called nu in doc
        """
        # *** store scale or inv_scale internally? ***
        self.scale = np.asarray(scale)
        self.df = df

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    @property
    def shape(self):
        return self.scale.shape

    @property
    def inv_scale(self):
        return np.linalg.inv(self.scale)

    @property
    def mean(self):
        """E{self}"""
        return self.df * self.scale

    @property
    def mean_inv(self):
        """E{ inv(self) },
        inv(self) = a random matrix with inverse-Wishart distribution
        """
        d = self.shape[0]
        if self.df > d + 1:
            return self.inv_scale / (self.df - d - 1)
        else:  # mean_inv is undefined
            return np.full_like(self.scale, np.nan)

    def mode_inv(self):
        """Mode of inv(self) = most likely corresponding covariance matrix
        when self is the precision matrix of a Gaussian random vector.
        """
        return self.inv_scale / (self.df + self.scale.shape[0] + 1)

    @property
    def mean_log_det(self):
        """E{ ln |self| }
        (self is positive definite, so sign of determinant is always = 1.)
        Needed for mean_logpdf of multivariate Gaussian random variable.
        """
        d = self.scale.shape[0]
        (s, log_det_scale) = np.linalg.slogdet(self.scale)
        if s <= 0:
            raise RuntimeError('Wishart scale matrix must be positive definite')
        return multi_psi(self.df / 2, d) + d * np.log(2) +  log_det_scale

    def logpdf(self, x):  # not needed in ItemResponseCalc, only for test
        """ln pdf(x | self), using scipy.stats.wishart
        :param x: 2D or 3D array or array-like list
        :return: lp = scalar lp = ln pdf(x | self)
            lp.shape == x.shape[:-2],
            i.e., == () if x is 2D, == x.shape[0] if x is 3D.

        NOTE: scipy.stats.wishart can only calculate logpdf for matrices stacked along 3rd axis,
            NOT first axis, like samples from rvs
        """
        # ***** first transpose in case 3D ? ***************
        # ***** OR implement explicitly, using np stacking convention ?
        return wishart.logpdf(x, scale=self.scale, df=self.df)

    def rvs(self, size=1):
        """Samples from Wishart distribution
        :param size: scalar integer number of sample matrices
        :return: S = 3D array
            S[..., :, :] = ...-th random precision matrix
            S.shape == (*size, *self.scale.shape)
        """
        # *** numpy.random can not yet sample a Wishart distribution
        # scipy.wishart.rvs may generate singular samples if self.df < self.shape[0]
        return wishart.rvs(size=size, scale=self.scale, df=self.df)

    # ------------------------------------------- VI update:
    # def adapt(self, groups, global_theta, prior):  # non needed, over-ridden anyway
    #     """Update distribution parameters using trait data from all groups and global distributions.
    #     :param groups: iterable of ir_group.ItemResponseGroup instances
    #     :param global_theta: ir_global.GlobalTheta instance
    #     :param prior: object of same class as self
    #     :return: - KLdiv(self || prior)
    #     Result: updated internal parameters of self
    #
    #     Method: doc sec:PosteriorLambda, eqs eq:PosteriorNu, eq:PosteriorV
    #     """
    #     n_s = 0     # total number of respondents
    #     cov = np.zeros(self.scale.shape)  # sum inter-individual covariance
    #     for g in groups:
    #         n_s += g.n_subjects
    #         cov += g.sum_theta_theta()
    #         mean_theta_g = g.theta_pop.mu.loc
    #         cov -= g.mu.learned_weight * mean_theta_g[:, None] * mean_theta_g   # eq:PosteriorV
    #     cov -= global_theta.mu.sum_square_mean()  # eq:PosteriorV
    #     self.scale = np.linalg.inv(prior.inv_scale + cov)   # eq:PosteriorV
    #     self.df = prior.df + n_s    # eqs eq:PosteriorNu
    #     return - self.relative_entropy(prior)

    def relative_entropy(self, othr):
        """Kullback-Leibler divergence between self and othr
        :param othr: single instance of same class as self
        :return: scalar KLdiv(q || p) = E_q{ln q(x) / p(x)},
            where q = self and p = othr

        Method: Using notation D = self.shape[0], V = self.scale,
        KLdiv = (q.df - p.df) * E{log_det(self)} / 2
            - (q.df - p.df) * ln(2) * D / 2
            + (p.df * log_det(p.V) - q.df * log_det(q.V)) / 2
            + trace[(inv(p.V) - inv(q.V)) * E{self}] / 2
            + gammaln_D(p.df/2) - gammaln_D(q.df/2)
        where
            E{log_det(self)} = psi_D(q.df/2) + D * ln(2) + log_det(q.V)
            E{self} = q.df * q.V
        = (q.df - p.df) * psi_D(q.df/2) / 2
            + p.df * (log_det(p.V) - log_det(q.V)) / 2
            + trace[(inv(p.V) - inv(q.V)) q.V] * q.df / 2
            + gammaln_D(p.df/2) - gammaln_D(q.df/2)
        """
        d = self.scale.shape[0]
        # Vpinv_Vq_dot = np.dot(othr.inv_scale, self.scale)
        Vpinv_Vq = np.linalg.solve(othr.scale, self.scale)
        (s, log_det_VV) = np.linalg.slogdet(Vpinv_Vq)
        return (0.5 * ((self.df - othr.df) * multi_psi(self.df/2, d)
                       - othr.df * s * log_det_VV
                       + (np.trace(Vpinv_Vq) - d) * self.df
                       )
                + multigammaln(othr.df/2, d) - multigammaln(self.df/2, d)
                )


# ------------------------------------------------ help function:
def multi_psi(a, d):
    """Multivariate psi function
    = first derivative of scipy.special.multigammaln(a, d) w.r.t a
    :param a: scalar ONLY
    :param d: scalar dimensionality
    :return: scalar
        = sum( psi( a + (1-j)/2) for j = 1,...,d ) (Wikipedia)
        = sum( psi( a - (j-1)/2) for j = 1,...,d )
        = sum( psi( a - j/2) for j = 0,...,d-1 )
    """
    return np.sum(psi(a - np.arange(d) / 2))


# ------------------------------------------------------------- Module TEST:

if __name__ == '__main__':

    # --------------------------- Testing WishartRV:
    nx = 3
    m = np.zeros(nx)
    mean_prec1 = np.array(np.arange(1, nx + 1))
    mean_prec2 = np.diag(mean_prec1)

    prec2 = WishartRV(df=nx + 1, scale=mean_prec2 + 1.)
    prec2_prior = WishartRV(df=nx - 1 + 0.1, scale=mean_prec2)
    print('KLdiv = ', prec2.relative_entropy(prec2_prior))

    x = prec2.rvs(size=5)
    print('x = prec2.rvs(3) = ', x)
    # *** NOTE: MUST transpose for scipy.wishart.logpdf to work
    x_t = x.transpose((1, 2, 0))
    print('logpdf(x) = ', prec2.logpdf(x_t))





"""This module defines help classes to calculate response thresholds
from given model parameters.

*** Classes:
ThresholdsFree: allowing all response thresholds freely adaptable.
ThresholdsMidFixed: forcing one mid-range threshold -> zero, other thresholds free

*** Version history:
* Version 1.0.0:
2023-07-14, copied EmaCalc -> ItemResponseCalc

2023-06-05, Fix to prevent numerical overflow / underflow in extreme cases,
            and general code cleanup.
            New module functions mapped_tau(w), d_mapped_tau(w), mapped_tau_inv(tau),
            and related changes to Thresholdxxx methods.
"""
import numpy as np
from scipy.special import logit, expit, softmax
import logging

logger = logging.getLogger(__name__)

W_EPSILON = np.sqrt(np.finfo(float).eps)
# = additive constant in mapped_tau(w) to prevent too small response interval widths.
# NOTE: ema_latent uses interval (tau_low - theta, tau_high - theta) for probability calculation,
# so we must ensure that this interval is non-zero also for very big theta like 1 / W_EPSILON
# Thus, (1 / W_EPSILON + W_EPSILON) - 1 / W_EPSILON must be > 0,
# and (1 + W_EPSILON**2) - 1 > 0.


# -------------------------------------------------------------
class ThresholdsFree:
    """All internal thresholds freely variable, as specified by model parameter vector eta,
    with number of parameters == number of internal thresholds == M - 1, where
    M = number of response categories
    """
    @staticmethod
    def n_param(n_categories):
        return n_categories - 1

    @staticmethod
    def tau(eta):
        """Mapping given log-category-width parameters to response thresholds.
        :param eta: = array with
            eta[..., m] = ...-th sample of parameter defining
                non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
            eta[..., M] assumed fixed == 0, NOT included in input argument
        :return: tau = threshold array, incl. all elements in [-inf, +inf]
            (tau[..., m], tau[..., m+1]) = (LOWER, UPPER) limits for m-th ordinal response interval
            tau[..., 0] ==  - np.inf
            tau[..., -1] == + np.inf
            tau.ndim == eta.ndim; tau.shape[-1] == eta.shape[-1] + 2
        """
        z_shape = (*eta.shape[:-1], 1)
        zeta = np.concatenate((eta, np.zeros(z_shape)), axis=-1)
        return mapped_tau(softmax(zeta, axis=-1))

    @staticmethod
    def d_tau(eta):
        """Jacobian of thresholds with respect to eta
        :param eta: = 1D or 2D array with
            eta[..., m] = ...-th sample of parameter defining
                 non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == M - 1, where M == number of response intervals.
            eta[..., M] assumed fixed == 0, NOT given as input
        :return: 2D or 3D array d_tau, with
            d_tau[..., m, i] = d tau[..., m] / d eta[..., i]; m = 0,..., M; i = 0, ..., M-1
                 where (tau[s, m], tau[s, m+1] = (LOWER, UPPER) limits of m-th response interval
            d_tau[..., 0, :] = d_tau[..., -1, :] = 0., for extreme limits at +-inf
            d_tau.ndim == eta.ndim + 1; d_tau.shape[-2:] == (M+1, M-1)
        """
        z_shape = (*eta.shape[:-1], 1)
        zeta = np.concatenate((eta, np.zeros(z_shape)), axis=-1)
        w = softmax(zeta, axis=-1)
        return jac_mapped_tau(w) @ _jac_softmax(w)[..., :-1]

    @staticmethod
    def tau_inv(tau):
        """Inverse of self.tau(eta)
        :param tau: 1D or 2D array of thresholds, NOT including extremes at -inf, + inf
        :return: eta, such that self.tau(eta) == tau
        """
        w = mapped_tau_inv(tau)
        # = normalized interval widths
        # w.shape[-1] == tau.shape[-1] + 1
        eta = np.log(w)
        eta -= eta[..., -1:]
        return eta[..., :-1]

# -------------------------------------------------------------
class ThresholdsMidFixed:
    """One mid-range threshold fixed == 0.,
    other thresholds mapped from parameter array eta,
    with number of parameters == M - 2, where
    M = number of response categories
    """
    @staticmethod
    def n_param(n_categories):
        return n_categories - 2

    @staticmethod
    def tau(eta):
        """Mapping given log-category-width parameters to response thresholds.
        :param eta: = array with
            eta[..., m - 1] = ...-th sample of parameter defining
                log(non-normalized width) of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
        :return: tau = 1D or 2D array, INCL. all elements in [-inf, +inf]
            (tau[..., m], tau[..., m+1]) = (LOWER, UPPER) limits for m-th ordinal response interval
            tau[..., 0] ==  - np.inf
            tau[..., -1] == + np.inf
            tau.ndim == eta.ndim; tau.shape[-1] == eta.shape[-1] + 2
        """
        pad = np.zeros((*eta.shape[:-1], 1))
        zeta = np.concatenate((pad, eta, pad),
                              axis=-1)
        n_half = zeta.shape[-1] // 2
        w = np.concatenate((softmax(zeta[..., :n_half], axis=-1),
                            softmax(zeta[..., n_half:], axis=-1)), axis=-1)
        return mapped_tau(w)

    @staticmethod
    def d_tau(eta):
        """Jacobian of tau(eta) w.r.t. eta
        :param eta: = 1D or 2D array with
            eta[..., m - 1] = ...-th sample of parameter defining
                non-normalized width of m-th interval in mapped domain [0, 1].
            eta.shape[-1] == self.n_param(M), with M == number of response categories.
        :return: 2D or 3D array d_tau, with
            d_tau[..., m, i] = d tau[..., m] / d eta[..., i]; m = 0,..., M; i = 0, ..., M-1
                 where (tau[s, m], tau[s, m+1] = (LOWER, UPPER) limits of m-th response interval
            d_tau[..., 0, :] = d_tau[..., -1, :] = 0., for extreme limits at +-inf
            d_tau.ndim == eta.ndim + 1; d_tau.shape[-2:] == (M+1, M-2)
        """
        pad = np.zeros((*eta.shape[:-1], 1))
        zeta = np.concatenate((pad, eta, pad),
                              axis=-1)
        n_half = zeta.shape[-1] // 2
        w1 = softmax(zeta[..., :n_half], axis=-1)    # normalized lower half
        w2 = softmax(zeta[..., n_half:], axis=-1)    # normalized upper half
        dw_dzeta = np.zeros((*zeta.shape, zeta.shape[-1]))
        dw_dzeta[..., :n_half, :n_half] = _jac_softmax(w1)
        dw_dzeta[..., n_half:, n_half:] = _jac_softmax(w2)
        # dw_deta = dw_dzeta[..., 1:-1]  # EXCL fixed pad elements
        dtau_dw = jac_mapped_tau(np.concatenate((w1, w2), axis=-1))
        return dtau_dw @ dw_dzeta[..., 1:-1]

    @staticmethod
    def tau_inv(tau):
        """Inverse of self.tau(eta)
        :param tau: 1D or 2D array of thresholds, NOT including extremes at -inf, + inf
        :return: eta, such that self.tau(eta) == tau
            after adjustment of middle tau element -> zero
        """
        n_half = (tau.shape[-1] + 1) // 2
        # = index of mid tau == index of corresponding w
        mid = tau[..., [n_half - 1]]
        w = mapped_tau_inv(tau - mid)
        eta = np.log(w)
        eta[..., :n_half] -= eta[..., 0:1]
        eta[..., n_half:] -= eta[..., -1:]
        return eta[..., 1:-1]



# ----------------------------- general module functions

# NOTE: scipy logit, expit are NOT symmetric around mid-point for extreme arguments
# logit(expit(37.)) = 37.; logit(expit(38.)) = +inf; logit(expit(-38.)) = -38.
# logit(expit(-709.)) = -709.; logit(expit(-710.)) = -inf
# *** Now should be sufficiently protected by W_EPSILON in mapped_tau() function

def mapped_tau(w):
    """Response thresholds from given un-normalized interval widths
    :param w: array of row vector(s) with POSITIVE width parameters
        normalized with sum == 1. for ThresholdsFree, or 2. for ThresholdsMidFixed
        w.shape[-1] == number of ordinal response levels.
    :return: tau = corresponding threshold values, INCL. extreme -inf, ..., +inf
        tau.shape[-1] == w.shape[-1] + 1
    """
    w = w + W_EPSILON   # ensure no interval width gets truncated -> 0.
    cum_w = np.cumsum(w, axis=-1)
    z_shape = (*cum_w.shape[:-1], 1)
    cum_w = np.concatenate((np.zeros(z_shape), cum_w),
                           axis=-1)  # include cum_w[..., 0] = 0.
    # cw = cum_w[..., 1:-1]  # EXCL extreme == zero
    # sw = cum_w[..., -1:]
    # tau = logit(cw / sw) = np.log(cw) - np.log(sw - cw) EXCL extreme limits at -inf, +inf
    # + W_EPSILON HERE, TOO ?
    # ***** Check: close tau values + 1000 might get truncated to ZERO interval width
    tau = logit(cum_w / cum_w[..., -1:])  # INCL extreme at -inf, +inf
    # *** interval of (tau - theta) might get -> 0 ?
    d = np.diff(tau - 10000., axis=-1)
    d_warn = d == 0.
    if np.any(d_warn):
        logger.warning(f'mapped_tau: num. underflow: diff(tau) = \n' +
                       np.array_str(d[d_warn]) +
                       f'\nw =\n' + np.array_str(w[d_warn]))
        if tau.ndim > 1:
            d_warn = np.any(d_warn, axis=-1)
            logger.warning(f'mapped_tau: tau = \n'+
                           np.array_str(tau[d_warn]))
    return tau


def jac_mapped_tau(w):
    """Jacobian of mapped_tau(w) with respect to w
    :param w: 1D or 2D array with POSITIVE width parameters >= ETA_W_EPSILON
    :return: 2D or 3D array d_tau, with
        d_tau[..., m, i] = d tau[..., m] / d w[..., i],
            for m = 0,..., M + 1; i = 0, ..., M - 1
            where (tau[s, m], tau[s, m+1]) = (LOWER, UPPER) limits of m-th response interval
        d_tau[..., 0, :] = d_tau[..., -1, :] == 0., for extreme limits at +-inf
        d_tau.shape == (N, M+1, M); N = n_samples; M = n response categories
    """
    w = w + W_EPSILON
    nw = w.shape[-1]
    cum_w = np.cumsum(w, axis=-1)
    cw = cum_w[..., :-1, np.newaxis]  # only inner limits
    sw = cum_w[..., -1:, np.newaxis]
    # dcw_dw[..., m, i] = dcw[..., m, 0] / dw[..., i]  = 1. if i <= m else 0.
    dcw_dw = np.tril(np.ones((nw - 1, nw), dtype=int))  # dtype = Boolean ?
    # tau[..., m+1] = ln cw[..., m, 0]  - ln (sw[..., 0, 0] - cw[..., m, 0])
    dtau_dw = dcw_dw / cw - (1 - dcw_dw) / (sw - cw)
    pad = np.zeros((*dtau_dw.shape[:-2], 1, dtau_dw.shape[-1]))
    return np.concatenate((pad,
                           dtau_dw,
                           pad), axis=-2)


def mapped_tau_inv(tau):
    """Approximate inverse of mapped_tau(w)[..., 1:-1]
    :param tau: 1D or 2D array with response thresholds, EXCEPT extremes at +-inf,
        i.e., all tau elements in (-inf, +inf),
        tau[..., m] = UPPER limit for m-th interval,
            = LOWER limit for the (m+1)-th interval
        tau.shape[-1] == number of response intervals - 1
    :return: w: 1D or 2D array, such that
        mapped_tau(w) approx == tau (except for added W_EPSILON)
    """
    y = expit(tau)
    cat_shape = (*y.shape[:-1], 1)
    y = np.concatenate((np.zeros(cat_shape),
                        y,
                        np.ones(cat_shape)), axis=-1)
    # = including extreme limits at 0 and 1
    return np.diff(y, axis=-1)


# ----------------------------- local module help functions
def _jac_softmax(w):
    """Jacobian of w = softmax(eta, axis=-1) w.r.t eta,
    calculated as a function of w,
    because w must be already calculated by caller anyway
    :param w: = array of ROW vectors with normalized width values = softmax(eta, axis=-1)
    :return: array dw, with
        dw[..., i, m] = d softmax(eta, axis=-1)[..., i] / d eta[..., m]
        dw.shape == (*w.shape, w.shape[-1]) == (*eta.shape, eta.shape[-1])
    """
    n = w.shape[-1]
    dw = - w[..., :, None] * w[..., None, :]    # dw[..., i, m] = w[..., i] * w[..., m]
    dw[..., range(n), range(n)] += w            # dw[..., i, i] += w[..., i]
    return dw


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    from scipy.optimize import approx_fprime, check_grad

    n_categories = 4  # 2, 3, 5
    # *** extreme values:
    eta_extreme = -100000.  # -1000.

    print('\n*** Testing mapped_width = softmax ***')
    eta = np.array([-5., 0., 5, 10.])
    eta = np.zeros(n_categories)
    eta[1] = eta_extreme
    # eta = - eta_extreme * np.ones(n_categories)
    # all these cases work OK now

    w = softmax(eta)
    print(f'softmax({eta}, axis=-1) = ', w)
    print(f'dsoftmax_deta({eta}) = \n', _jac_softmax(w))
    print(f'd sum_softmax({eta} = \n', np.sum(_jac_softmax(w), axis=0))
    for i in range(len(eta)):
        def fun(eta):
            return softmax(eta)[i]
        def jac(eta):
            w = softmax(eta)
            return _jac_softmax(w)[i]
        print('approx gradient = ', approx_fprime(eta, fun, epsilon=1e-6))
        print('exact  gradient = ', jac(eta))
        err = check_grad(fun, jac, eta, epsilon=1e-6)
        print('check_grad err = ', err)

    print('\n*** Testing mapped_tau ***')
    w = 1. + np.arange(n_categories)
    # w /= np.sum(w)  # normalized
    tau = mapped_tau(w)
    print(f'mapped_tau({w} = ', tau)
    tau_inv = mapped_tau_inv(tau[1:-1])
    print(f'mapped_tau_inv({tau[1:-1]} = ', tau_inv)
    print('Max(mapped_tau_inv - normalized(w)) = ', max(tau_inv - w / np.sum(w)))

    for i in range(1, len(tau)-1):
        def fun(w):
            return mapped_tau(w)[i]
        def jac(w):
            return jac_mapped_tau(w)[i]

        print('approx gradient = ', approx_fprime(w, fun, epsilon=1e-6))
        print('exact  gradient = ', jac(w))
        err = check_grad(fun, jac, w, epsilon=1e-6)
        print('check_grad err = ', err)

    n_samples = 1
    print(f'\nn_categories = {n_categories}')

    for thr in [ThresholdsFree, ThresholdsMidFixed]: # , Thresholds
        eta = np.zeros(thr.n_param(n_categories))

        # eta -= eta_extreme
        if len(eta) > 0:
            eta[0] = eta_extreme
            eta[-1] = - eta_extreme

        print(f'\n*** Testing {thr.__name__}.tau ***')

        tau = thr.tau(eta)
        print(f'eta = ', eta)
        # w = mapped_width(eta)
        # print(f'mapped_width({eta} = ', w)
        print(f'tau({eta}) = ', tau)
        print(f'diff(tau) = ', np.diff(tau, axis=-1))
        # print(f'tau_inv(tau[..., 1:-1] = ', thr.tau_inv(tau[..., 1:-1]))
        # print(f'tau(tau_inv(tau[..., 1:-1]) = ', thr.tau(thr.tau_inv(tau[..., 1:-1])))
        print(f'd_tau({eta}) = ', thr.d_tau(eta))

        def fun(eta):
            eta = np.tile(eta, (n_samples, 1))
            return thr.tau(eta)[0, limit]

        def jac(eta):
            eta = np.tile(eta, (n_samples, 1))
            return thr.d_tau(eta)[0, limit]

        if eta.ndim > 1:
            eta_test = eta[0]  # must be 1D vector for gradient test
        else:
            eta_test = eta

        for limit in range(1, n_categories):
            print(f'\n*** Testing Jacobian {thr.__name__}.d_tau[..., {limit}, :] ***')

            print(f'tau({eta_test}) = {thr.tau(eta_test)}')

            print('approx gradient = ', approx_fprime(eta_test, fun, epsilon=1e-6))
            print('exact  gradient = ', jac(eta_test))
            err = check_grad(fun, jac, eta_test, epsilon=1e-6)
            print('check_grad err = ', err)




"""This module implements a random variable with gamma distribution,
or an array of independent such variables.

Extends the scipy.stats implementation by including Bayesian learning.
"""
import numpy as np
from scipy.special import gammaln, psi

class GammaRV:
    """Implements a vector of INDEPENDENT gamma-distributed random variables.
    The probability density function is
    p(Lambda) = prod_d C_d Lambda_d^(a_d - 1) exp(- b_d Lambda_d), i.e., a gamma density,
        where
        a = scalar (or 1D array) shape parameters
        b = 1D array of inverse-scale parameters
        Lambda.shape == b.shape
        a and b must have broadcast-compatible shapes
        C_d = b_d^a / Gamma(a) is the normalization factor
    """
    def __init__(self, a=0., b=1.):
        """
        :param a: scalar or 1D array-like shape parameter(s)
        :param b: 1D array-like with inverse scale parameter(s)
        """
        # assert np.isscalar(a), 'shape parameter should be scalar for IRT usage'
        try:
            a = np.array(a)
            b = np.array(b)
            test = a / b
        except ValueError as e:
            raise RuntimeError('a and b parameters must be broadcast-compatible and > 0. ' + str(e))
        self.a = a
        self.b = b

    def __repr__(self):
        return self.__class__.__name__ + f'(a= {repr(self.a)}, b= {repr(self.b)})'

    @property
    def size(self):
        return self.mean.size

    @property
    def shape(self):
        return self.mean.shape

    @property
    def scale(self):
        return 1./self.b

    @property
    def inv_scale(self):
        return self.b

    @property
    def mean(self):
        """E{self}"""
        return self.a / self.b

    @property
    def mean_inv(self):
        """E{ inv(self) }, where
        inv(self) has an inverse-gamma distribution
        """
        if self.a.ndim > 0:
            m = self.b / np.maximum(self.a - 1, np.finfo(float).eps)
            m[self.a <= 1.] = np.nan
            return m
        elif self.a <= 1.:
            return np.full_like(self.b, np.nan)
        else:
            return self.b / (self.a - 1)

    def mode_inv(self):
        """mode{ inv(self) }, where
        inv(self) has an inverse-gamma distribution
        """
        return self.b / (self.a + 1.)

    @property
    def mean_log(self):
        """E{ ln self } element-wise"""
        return psi(self.a) - np.log(self.b)

    def logpdf(self, x):
        """ln pdf(x | self)
        :param x: array or array-like list
            x.shape[-1:] == self.mean.shape
        :return: lp = scalar or array, with
            lp[...] = ln pdf(x[..., :] | self)
            lp.shape == x.shape[:-1]
        """
        bx = self.b * np.asarray(x)
        return np.sum((self.a - 1.) * np.log(bx) - bx
                      + np.log(self.b) - gammaln(self.a),
                      axis=-1)

    # def adapt(self, x2, w, prior, new_mean_2):  # overridden anyway in ItemResponseCalc
    #     """Update distribution parameters using observed data and prior.
    #     :param x2: 2D array or array-like list of squared samples
    #         for vectors assumed drawn from distribution with precision == self.
    #     :param w: 1D array with sample weights
    #     :param prior: object of same class as self
    #     :param new_mean_2: weighted difference = nu * new_loc^2 - nu' * prior_loc**2
    #         where nu is the new sum-weight and nu' is the prior sum-weight
    #     :return: None
    #
    #     Result: updated internal parameters of self
    #     Method: Leijon EmaCalc report: sec:GaussGammaUpdate
    #     """
    #     self.a = prior.a + np.sum(w) / 2
    #     # eq:GammaUpdateA
    #     self.b = prior.b + (np.dot(w, x2) - new_mean_2) / 2
    #     # eq:GammaUpdateB

    def relative_entropy(q, p):
        """Kullback-Leibler divergence between PrecisionRV q and p,
        :param p: another instance of same class as self = q
        :return: scalar KLdiv( q || p ) = E{ ln q(x)/p(x) }_q

        Arne Leijon, 2018-07-07 copied from gamma.py 2015-10-16, checked 2023-07-25
        """
        pb_div_qb = p.b / q.b
        return np.sum(gammaln(p.a) - gammaln(q.a)
                      - p.a * np.log(pb_div_qb)
                      + (q.a - p.a) * psi(q.a)
                      - q.a * (1. - pb_div_qb)
                      )




"""This module defines data structures and functions to display analysis results
given an OrdinalItemResponseModel instance,
learned from one or several data sets with ordinal responses to a questionnaire.

Results are shown as figures and tables.
Figures can be saved in pdf, eps, or other formats, as specified in ir_display_format.
Tables are saved in LaTeX tabular format OR in tab-delimited text files.
Thus, both figures and tables can be easily imported into a LaTeX document
or other word-processing document.

*** Main Class:

ItemResponseDisplaySet = a structured container for all display results
Each display element can be accessed and modified by the user, before saving.

*** Usage Example: see template script run_irt

*** Version History:
* Version 1.0.0:
2023-08-30, code cleanup
2023-08-13, adapted to modified ir_model, other minor updates

* Version 0.6:
2023-07-06: Table and figure names set when created. File formats defined in save methods.
2023-07-06: table results as pandas.DataFrame objects.

2020-06-04, first version for ItemResponseCalc standalone package
"""
# **** Allow scaling here, for unity inter-individual trait variance.
# **** display individual results ???

import numpy as np
from collections import Counter
from pathlib import Path

import logging

from samppy.credibility import cred_diff

from . import ir_display_format as fmt

# -------- module-global variables, that may be changed by user:
FMT = {'percentiles': [5., 50., 95.],
       'credibility_levels': [.99, .95, .9, .8, .7],  # MAX 5 values
       'population': 'Population',
       'trait': 'Trait',
       'mean': 'Mean',
       'random': 'Random',
       'individual': 'Individual',
       'equivalent': 'Equivalent',
       'item': 'Item',
       'response': 'Response',
       'thresholds': 'Thresholds',
       'probability': 'Prob.',
       'rating': 'Rating',
       'global': 'Global',
       'correlation': 'Correlation',
       'credibility': 'Credibility',
       'quantile': 'Quantile',
       # 'standardize': False,  # Future? True -> standardized trait scales in all displays
       }
# = dict with format parameters that may be changed by user
# Any other parameters are handled by module ir_display_format

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # test

DEFAULT_FIGURE_FORMAT = 'pdf'
DEFAULT_TABLE_FORMAT = 'txt'


# ------------------------------------------------------------
def set_format_param(**kwargs):
    """Set / modify format parameters for this module
    :param kwargs: dict with any formatting variables
    :return: None
    """
    other_fmt = dict()
    for (k, v) in kwargs.items():
        k = k.lower()
        if k in FMT:
            FMT[k] = v
        else:
            other_fmt[k] = v
    if len(FMT['credibility_levels']) > 5:
        FMT['credibility_levels'] = FMT['credibility_levels'][:5]
    fmt.set_format_param(**other_fmt)  # all remaining user-defined parameters


# ---------------------------------------------------------- Main Display Classes:
class ItemResponseDisplaySet:
    """Root container for all displays of predictive trait results and item properties,
    from one OrdinalItemResponseModel object.
    All display elements can be saved as files within a selected directory three.
    The complete instance can also be serialized and dumped to a single pickle file,
    then re-loaded and re-saved, if any display object needs to be edited.
    """
    def __init__(self,
                 log_likelihood,
                 instrument,
                 traits):
        """
        :param log_likelihood: fmt.FigureRef with log-likelihood plot
        :param instrument: ItemSetDisplay object, showing
            response thresholds, and response_prob plot for each item
            (containing only FigureRef displays)
        :param traits: TraitDisplay object with predicted trait distribution
            for population mean, and for random individual in population.
            Traits for individual respondents are NOT (yet) displayed.
            (containing both FigureRef and ResultTable instances)
        """
        self.log_likelihood = log_likelihood
        self.instrument = instrument
        self.traits = traits

    def __repr__(self):
        skip = '\n\t'
        return (self.__class__.__name__ + '(' + skip +
                (',' + skip).join(f'{key}={repr(v)}'
                                  for (key, v) in vars(self).items()) +
                skip + ')')

    def save(self, dir_top,
             figure_format=None,
             table_format=None,
             **kwargs):
        """Save all displays in a directory tree
        :param dir_top: Path or string with top directory for all displays
        :param figure_format: (optional) single figure-file format string
        :param table_format: (optional) single table-file format string
        :param kwargs: (optional) additional parameters passed to ResultTable.save() method.
            NOTE: NO extra kwargs allowed for FigureRef.save() method!
        :return: None
        """
        dir_top = Path(dir_top)  # just in case
        if figure_format is None and table_format is None:
            figure_format = DEFAULT_FIGURE_FORMAT
            table_format = DEFAULT_TABLE_FORMAT
        # One of them may be None -> only the other type of data are saved.
        dir_top.mkdir(parents=True, exist_ok=True)  # just in case
        self.log_likelihood.save(dir_top, figure_format=figure_format)
        if self.instrument is not None:
            self.instrument.save(dir_top,
                                 figure_format=figure_format,
                                 table_format=table_format)
        if self.traits is not None:
            self.traits.save(dir_top,
                             figure_format=figure_format,
                             table_format=table_format,
                             **kwargs)

    @classmethod
    def show(cls, irm, mapping_item=None, **kwargs):
        """Create displays for all results from learned model,
        and store all display elements in a single cls instance.
        :param irm: a single learned OrdinalItemResponseModel instance
        :param mapping_item: index of item to be used for
            mapping non-linear mean rating, given scale trait
        :param: kwargs: (optional) any display formatting parameters
        :return: a single cls instance
        """
        set_format_param(**kwargs)
        logger.info(f'Creating displays for {len(irm.groups)} population(s): ' +
                    ', '.join(g.name for g in irm.groups))
        logger.info(fig_comments())
        logger.info(table_comments())
        return cls(log_likelihood=fmt.fig_log_likelihood(irm.log_prob,
                                                         irm.n_subjects,
                                                         color='r',
                                                         name='LearnedLL'),
                   instrument=ItemSetDisplay.show(irm),
                   traits=TraitDisplay.show(irm, mapping_item))


class ItemSetDisplay:
    """Container for all displays of item-related properties
    """
    def __init__(self,
                 response_freq,
                 items,
                 thresholds=None,
                 fisher_info=None):
        """
        :param response_freq: FigureRef with relative response frequencies for all items
        :param items: dict with elements (item_id, item_display),
             where item_display = ItemDisplay object, containing FigureRef objects
        :param thresholds: TableRef with response thresholds for all items (and all groups)
        :param fisher_info: dict with elements (trait_id, fisher_fig)
            where fisher_fig is a FigureRef with Item Information Function plots
            for items related to the same trait.
        """
        self.response_freq = response_freq
        self.items = items
        self.thresholds = thresholds
        if fisher_info is None:
            fisher_info = dict()
        self.fisher_info = fisher_info

    def save(self, top_path, figure_format, table_format):
        """
        :param top_path: path where plots are saved
        :param figure_format: file format for saved figures
        :param table_format: file format for saved tables
        :return: None
        """
        items_path = top_path / 'items'
        items_path.mkdir(parents=True, exist_ok=True)
        if self.response_freq is not None:
            self.response_freq.save(items_path, figure_format)
        for item in self.items.values():
            item.save(items_path, figure_format)
        if self.thresholds is not None:
            self.thresholds.save(items_path, table_format)
        for f_info in self.fisher_info.values():
            f_info.save(items_path, figure_format)

    @classmethod
    def show(cls, irm):
        """
        :param irm: OrdinalItemResponseModel instance
        :return: a cls instance
        """
        (t_min, t_max) = irm.theta_range()
        t = np.linspace(t_min, t_max, 100)
        # same trait range across all items
        g_count = irm.item_response_count()
        # = dict with elements (group_key, g_count), where
        # g_count = list of Counter objects, one for each item
        total_count = [sum((g_c[i] for g_c in g_count.values()), Counter())
                       for i in range(irm.n_items)]
        # *** use item_id instead of item_index ********************
        response_prob = irm.item_response_prob(t)
        # = list with 1D response vectors, one for each item
        item_id = np.array(irm.item_labels())  # to allow boolean indexing
        thr = irm.item_thresholds(item=FMT['item'], response=FMT['response'])
        thr = thr.groupby([FMT['item'], FMT['response']]).quantile(np.array(FMT['percentiles']) / 100.)
        thr.index.set_names(FMT['quantile'], level=-1, inplace=True)  # ***** -> irm.item_thresholds_quantiles?
        # = pd.DataFrame with threshold quantiles for all items and all groups
        # with row MultiIndex (item, response, quantile), and one column for each group
        f_t = irm.item_fisher_information(t)
        item_trait_map = irm.item_trait_map()
        # ************** weighted average instead of hard selection ? ************************
        trait_labels = fmt.make_trait_labels(item_trait_map)  # *** -> irm ?
        fisher_figs = {trait_id: fmt.fig_fisher_information(t, f_t[trait_items, :],
                                                            item_labels=item_id[trait_items],
                                                            x_label=FMT['trait'],
                                                            y_label='Fisher Information',
                                                            name=trait_id+'-Fisher')
                       for (trait_id, trait_items) in zip(trait_labels,
                                                          item_trait_map.T)
                       }
        return cls(response_freq=fmt.fig_response_freq(total_count,
                                                       name='Response_freq'),
                   items={id_i: ItemDisplay.show(r_prob_i,
                                                 trait=t,
                                                 tau_global=tau_i[1:-1],
                                                 item_id=id_i,
                                                 tau_group=thr.loc[id_i])
                          for (r_prob_i, id_i, tau_i) in zip(response_prob,
                                                             item_id,
                                                             irm.item_thresholds_global())
                          },
                   thresholds=fmt.ResultTable(thr, name='ItemThresholds'),
                   fisher_info=fisher_figs
                   )


class ItemDisplay:
    """Container for all displays related to ONE item
    """
    def __init__(self,
                 response_freq=None,
                 response_prob=None,
                 mean_response=None,
                 thresholds=None):
        """
        :param response_freq: FigureRef with empirical response freq
        :param response_prob: FigureRef with predicted response probability vs. trait,
            like 'Category Characteristic Curves' in the Rasch tradition.
        :param mean_response: FigureRef with expected response vs. trait
            (*** and ? empirical response range vs trait ? *****)
            NOTE: NOT valid, as ordinal responses are NOT on interval scale
            *** SKIPPING this display variant for now...
        :param thresholds: FigureRef with threshold quartiles for all groups
        """
        self.response_freq = response_freq  # ****** NOT USED for now ***
        self.response_prob = response_prob
        self.mean_response = mean_response  # ****** NOT USED for now ***
        self.thresholds = thresholds

    def save(self, save_path, figure_format):
        """
        :param save_path: path where sub-objects are saved
        :param figure_format: file format of saved figures
        :return:
        """
        if figure_format is not None:
            if self.response_freq is not None:
                self.response_freq.save(save_path, figure_format)
            if self.response_prob is not None:
                self.response_prob.save(save_path, figure_format)
            if self.mean_response is not None:
                self.mean_response.save(save_path, figure_format)
            if self.thresholds is not None:
                self.thresholds.save(save_path, figure_format)

    @classmethod
    def show(cls, response_prob, trait, item_id, tau_global=None, tau_group=None):
        """Generate displays for given item
        :param response_prob: 2D array with response prob.mass
            response_prob[n, l] = prob.mass vector, for given trait[n] and l-th ordinal response
        :param trait: 1D array with dense trait values for plot
        :param item_id: string for plot lobel
        :param tau_global: 1D array with global mean thresholds for this item
            to be included in response-prob. plots
        :param tau_group: pd.DataFrame with threshold quantiles for this item
            separated by group, displayed in a separate figure
        :return: a cls instance
        """
        # r_mean = 1. + np.dot(ord_prob, np.arange(item_scale.n_response_levels))
        # response_freq=fmt.fig_response_freq(response_count),
        if tau_group.shape[-1] > 1:  # more than ONE population
            tau_fig = fmt.fig_thresholds(tau_group,
                                         name=item_id + FMT['thresholds'],
                                         x_label=FMT['population'],
                                         y_label=' '.join([FMT['response'],
                                                           FMT['thresholds']])
                                         )
        else:
            tau_fig = None
        return cls(response_prob=fmt.fig_response_prob(trait, response_prob,
                                                       name=item_id + FMT['probability'],
                                                       tau=tau_global,
                                                       x_label=FMT['trait'],
                                                       y_label=' '.join([FMT['response'],
                                                                         FMT['probability']])
                                                       ),
                   thresholds=tau_fig
                   )


class TraitDisplay:
    """Container for result displays of predicted trait values
    for all traits and groups,
    with layout depending on number of groups and traits.
    """
    def __init__(self, pop_mean, pop_ind,
                 corr_global=None,
                 corr_within=None,
                 group_diff=None):
        """
        :param pop_mean: FigureRef of population predictive mean traits
        :param pop_ind: FigureRef of population predictive traits for random individual
        :param corr_global: (optional) TableRef with global predictive trait correlations
        :param corr_within: (optional) TableRef with predictive trait correlations
            within populations, i.e., excl. variance between populations.
        :param group_diff: (optional) GroupDiffDisplay object,
            iff we have more than one group (population)
       """
        self.pop_mean = pop_mean
        self.pop_ind = pop_ind
        self.corr_global = corr_global  # *** not used in v. 1.0.0 ***
        self.corr_within = corr_within
        self.group_diff = group_diff

    def save(self, path, figure_format, table_format, **kwargs):
        """
        :param path: to directory where all display objects are saved
        :param figure_format: file format for saved figures
        :param table_format: file format for saved tables
        :param kwargs: (optional) additional parameters to table save method
        :return:
        """
        if figure_format is not None:
            if self.pop_mean is not None:
                self.pop_mean.save(path, figure_format)
            if self.pop_ind is not None:
                self.pop_ind.save(path, figure_format)
        if self.corr_global is not None:
            self.corr_global.save(path, table_format, **kwargs)
        if self.corr_within is not None:
            self.corr_within.save(path, table_format, **kwargs)
        if self.group_diff is not None:
            for (trait_id, trait_diff) in self.group_diff.items():
                trait_diff.save(path / 'group_diff',
                                figure_format=figure_format,
                                table_format=table_format,
                                **kwargs)

    @classmethod
    def show(cls, irm, mapping_item=None):
        """
        :param irm: OrdinalItemResponseModel instance
        :param mapping_item: index of item to be used for non-linear scale mapping
            *** identify by ID string instead ? ***
        :return: a cls instance
        """
        # *** show percentiles as table, too ? ***
        def _fig_percentiles(perc, g_labels, t_labels, y_label, name=None):
            """Plot trait percentiles vs groups and/or traits,
            in style depending on n_groups and n_traits
            :param perc: 3D array with percentile trait values
                perc[g, p, t] = p-th percentile of t-th trait in g-th population
            :param g_labels: list of group label strings
                len(g_labels) == perc.shape[0]
            :param t_labels: list of trait label strings
                len[t_labels] == perc.shape[2]
            :param y_label: string with y-axis label
            :param name: string with file name for the fiture
            :param mapping_item: index of irm.items to be used for mapping
            :return: FigureRef instance
            """
            (n_groups, n_perc, n_traits) = perc.shape
            if n_traits > 1:  # plot perc vs traits, with populations as subcases
                f = fmt.fig_percentiles(perc,
                                        x_label=FMT['trait'],
                                        x_ticklabels=t_labels,
                                        case_labels=g_labels,
                                        name=name,
                                        y_label=y_label,
                                        x_offset=0.07)
            else:  # plot perc vs groups, with traits as subcases
                f = fmt.fig_percentiles(perc.transpose(2, 1, 0),
                                        x_label=FMT['population'],
                                        x_ticklabels=g_labels,
                                        case_labels=t_labels,
                                        name=name,
                                        y_label=y_label,
                                        x_offset=0.07)
            (y_min, y_max) = f.ax.get_ylim()
            y_min -= 0.1  # **** fix to make more space for legend
            f.ax.set_ylim(y_min, y_max)
            if mapping_item is not None:
                # make second corresponding scale with expected response ratings
                y = np.linspace(y_min, y_max, 100)
                # use mapping_item and its item_id
                y_mapped = 1. + irm.item_mean_ordinal(y)[mapping_item]
                item_id = f'Q{1+mapping_item}'  # *** get mapping_item as ID string ?
                f.mapped_y_axis(y, y_mapped,
                                y_label=(FMT['equivalent'] + ' ' +
                                         item_id + ' ' + FMT['rating'])
                                )
            return f
        # -----------------------------------------------------------
        if mapping_item < 0:
            mapping_item = irm.n_items + mapping_item # ensure positive index
        item_trait_map = irm.item_trait_map()
        trait_sel = np.any(item_trait_map, axis=0)
        # = boolean trait index for display, just in case irm has NOT been pruned
        # *** should display the non-pruned model?
        # *** NO, because only traits[trait_sel] are effectively used in the model.
        n_traits = np.sum(trait_sel)
        trait_labels = fmt.make_trait_labels(item_trait_map[:, trait_sel])
        group_labels = [g.name for g in irm.groups]  # ********* irm.groups.keys()
        global_mu = irm.global_pop.theta.predictive_mean()
        global_mu_rvs = global_mu.rvs(size=10000)
        pop_mean_samples = np.array([g.predictive_mean().rvs(global_mu_rvs)[:, trait_sel]
                                     for g in irm.groups])
        # pop_mean_samples[g, n, t] = n-th sample of t-th trait in g-th population
        # all subpopulations conditional on global_mu
        pop_mean_perc = np.percentile(pop_mean_samples,
                                      FMT['percentiles'],
                                      axis=1)
        # pop_mean_perc[p, g, t] = p-th percentile for t-th trait in g-th population
        pop_mean_fig = _fig_percentiles(pop_mean_perc.transpose(1, 0, 2),
                                        group_labels,
                                        trait_labels,
                                        y_label=' '.join([FMT['population'],
                                                          FMT['mean'],
                                                          FMT['trait']]),
                                        name=' '.join([FMT['trait'],
                                                       FMT['mean']]))
        pop_ind_samples = np.array([g.predictive_individual().rvs(global_mu_rvs)[:, trait_sel]
                                    for g in irm.groups])
        pop_ind_perc = np.percentile(pop_ind_samples,
                                      FMT['percentiles'],
                                      axis=1)
        # pop_ind_perc[p, g, t] = p-th percentile for t-th trait in g-th population
        pop_ind_fig = _fig_percentiles(np.array(pop_ind_perc).transpose(1, 0, 2),
                                       group_labels,
                                       trait_labels,
                                       y_label=' '.join([FMT['random'],
                                                         FMT['individual'],
                                                         FMT['trait']]),
                                       name=' '.join([FMT['trait'],
                                                      FMT['individual']]))

        c_global = None
        c_within = None
        if n_traits > 1:
            c_global = fmt.tab_correlation_matrix(irm.predictive_individual_cov(), item_trait_map,
                                                  name='_'.join((FMT['trait'],
                                                                 FMT['correlation'],
                                                                 FMT['global'])
                                                                ))
        g_diff = None
        if irm.n_groups > 1:
            # show differences between population means:
            g_diff = {trait_id: GroupDiffDisplay.show(t_samples,
                                                      group_labels,
                                                      trait_label=trait_id)
                      for (trait_id, t_samples) in zip(trait_labels,
                                                       pop_mean_samples.transpose(2, 0, 1))
                      }
            if n_traits > 1:
                g_diff['Mean_Trait'] = GroupDiffDisplay.show(np.mean(pop_mean_samples,
                                                                     axis=-1),
                                                             group_labels,
                                                             trait_label='Mean_Trait')

        return cls(pop_mean=pop_mean_fig,
                   pop_ind=pop_ind_fig,
                   corr_global=c_global,
                   corr_within=c_within,
                   group_diff=g_diff)


class GroupDiffDisplay:
    """Container for displays showing predictive credible differences
    between populations, for ONE selected trait, or MEAN across traits.
    """
    def __init__(self, fig=None, tab=None):
        """
        :param fig: FigureRef object
        :param tab: TableRef object
        """
        self.fig = fig
        self.tab = tab

    def save(self, path,
             figure_format,
             table_format,
             **kwargs):
        """
        :param path: path to directory for saved plots and tables
        :param figure_format: file format of saved figures
        :param table_format: file format of saved table
        :param kwargs: (optional) additional arguments to Table.save method
        :return: None
        """
        if self.fig is not None:
            self.fig.save(path, figure_format=figure_format)
        if self.tab is not None:
            self.tab.save(path, table_format=table_format, **kwargs)

    @classmethod
    def show(cls, group_samples, group_labels, trait_label):
        """Display credible differences between groups,
        given samples of predictive distribution
        :param group_samples: 2D array with
            group_samples[g, n] = n-th sample of selected trait for g-th group
        :param group_labels: list of string-valued group labels
            len(group_labels) == group_samples.shape[0]
        :param trait_label: string identifying the difference to be shown
        :return: a cls instance
        """
        # **** Sort groups for decreasing median values
        s_ind = np.argsort(np.median(group_samples, axis=1))[::-1]
        g_samples = group_samples[s_ind, :]
        g_labels = [group_labels[i] for i in s_ind]
        diff = cred_diff(g_samples,
                         diff_axis=0, sample_axis=1,
                         p_lim=np.amin(FMT['credibility_levels']))
        fig = fmt.fig_credible_diff(diff,
                                    diff_label=trait_label,
                                    x_labels=g_labels,
                                    x_label=FMT['population'],
                                    cred_levels=FMT['credibility_levels'])
        tab = fmt.tab_credible_diff(diff, diff_name=trait_label,
                                    cred_header=FMT['credibility'],
                                    x_labels=g_labels, x_label=FMT['population'])
        return cls(fig, tab)


# ------------------------------------------------ Result explanations
def fig_comments():
    """Generate figure explanations.
    :return: comment string
    """
    p_min = np.amin(FMT['percentiles'])
    p_max = np.amax(FMT['percentiles'])
    c = f"""Figure Explanations:

Figure files with a name like Population_Indiv_Trait.xxx or Population_Mean_Trait.xxx
display user-requested percentiles (markers) and credible intervals (vertical bars)
for the estimated trait distribution for a random individual and for the population mean.
The vertical bars show the range between {p_min:.1f}- and {p_max:.1f}- percentiles.

The displayed ranges include all uncertainty
caused both by real inter-individual perceptual differences,
and by the limited number of items and discrete response categories,
and (possibly) between-population differences in questionnaire interpretations.

Item-specific figures show the model-estimated probability for each response category
as a function of the latent trait parameter.
These plots are sometimes called 'Category Characteristic Curves' in Rasch tradition.

If there is more than one population,
item-specific figures show estimated response thresholds for all populations,
as median and range between {p_min:.1f}- and {p_max:.1f}- percentiles.
"""
    return c


def table_comments():
    c = """Table Explanations:

Files with names like someTrait.txt or, *.csv, tabulate all pairs of populations
for which the trait difference is JOINTLY credibly different,
with the tabulated credibility value.

The credibility value in each table row is the JOINT probability
for the pairs in the same row and all rows above it.
Thus, the joint probability values already account for multiple comparisons,
so no further adjustments are needed.

The same information is also saved in figure files with similar names, like someTrait.pdf,
showing all population pairs with jointly credible trait differences.

If several latent traits are needed to model the response data, 
table files with names like Trait_Correlation_xxx.xxx 
show the estimated correlation between traits. 
"""
    return c



"""This module implements a multivariate Student distribution of a random vector
with possibly DEPENDENT (correlated) vector elements,
or with INDEPENDENT vector elements,
and extends the scipy.stats implementations by including Bayesian learning.

This distribution is a predictive distribution for a multivariate Bayesian GaussianRV model.
The multivariate Student is obtained by integrating out the random variations of the precision of the Gaussian.

*** Reference:
M. Roth. On the multivariate t distribution.
Technical Report LiTH-ISY-R-3059, Linköping University, 2013.

*** Classes:
MultiStudentRV: a multivariate Student-t distribution allowing correlations between vector elements.
    NOTE: Even if elements are uncorrelated, i.e., the covariance matrix is diagonal,
    the elements are NOT independent, i.e.,
    the density function is NOT a product of scalar Student-t densities.

StudentRV: Student-t distribution for a vector with INDEPENDENT elements.

*** Version History:
* Version 1.0.0: new module within ItemResponseCalc
"""
import numpy as np
import logging
from scipy.special import gammaln  # , psi

RNG = np.random.default_rng()  # default random generator
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# ---------------------------------------------------------------------------
class MultiStudentRV:
    """Multivariate Student distribution of 1D random ROW vector
    with inter-dependent elements.
    The density function is (Roth, 2013; Wikipedia)
    f(x) = C [1 + (x - m)^T Sigma^{-1} (x - m) / df ]^{(df + D) / 2}
        where x and m are COLUMN vectors, D = len(x),
        df is degrees-of-freedom parameter,
        Sigma is symmetric positive-definite square matrix.
        (Covariance == Sigma * df / (df - 2), defined only for df > 2.)
        The normalization factor is
        C = Gamma((df + D) / 2) det(Sigma)^{-1/2) / Gamma(df / 2) df^{D/2) pi^{D/2)
    """
    def __init__(self, df,
                 loc=np.array(0.).reshape((1,)),
                 sigma=np.array(1.).reshape((1, 1))):
        """
        :param df: scalar degrees of freedom, MUST be > 0
        :param loc: 1D array or array-like list of location elements
        :param sigma: 2D symmetric positive semi-definite matrix
            (called shape matrix in scipy, scale matrix in Wikipedia)
            sigma.shape == (len(loc), len(loc))
        *** OR save only the scale (square root) matrix, s.t., sigma = scale @ scale.T ***
        """
        self.df = df
        self.loc = np.asarray(loc)
        self.sigma = np.asarray(sigma)

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    @property
    def shape(self):
        return self.loc.shape

    @property
    def size(self):
        return self.loc.size

    @property
    def mean(self):
        if self.df > 1.:
            return self.loc
        else:
            return np.full_like(self.loc, np.nan)

    @property
    def cov(self):
        """Covariance matrix"""
        if self.df > 2:
            return self.sigma * self.df / (self.df - 2.)
        else:
            return np.full_like(self.sigma, np.nan)

    def logpdf(self, x):
        """ln pdf(x | self)
        :param x: array or array-like list of sample vectors
            must be broadcast-compatible with self.loc
        :return: lp = scalar or array of logpdf values
            lp[...] = ln pdf[x[..., :] | self)
            lp.shape == x.shape[:-1]
        """
        raise NotImplementedError

    # def rvs_scipy(self, size=None, rng=RNG):
    #     """Generate array of random variates drawn from self
    #     :param size: integer or tuple with number of desired samples,
    #         where each sample.shape == self.shape
    #     :param rng: (optional) instance of np.random.Generator or subclass
    #     :return: x = array of samples
    #         x.shape == (*size, *self.shape)
    #     """
    #     # **** use np.random instead, with Gauss-gamma method (Roth, 2013)
    #     if size is None:
    #         result_shape = self.shape
    #     elif np.ndim(size) == 0:
    #         result_shape = (size,) + self.shape
    #     else:
    #         result_shape = size + self.shape
    #     x = multivariate_t.rvs(size=size,
    #                            df=self.df,
    #                            loc=self.loc,
    #                            shape=self.sigma)
    #     # multivariate_t squeezes result if loc.shape == (1,)
    #     if self.size == 1:  # must Un-sqeeze if x.shape != result_size
    #         x = x.reshape(result_shape)  # temp fix
    #     return x

    def rvs(self, size=None, rng=RNG):
        """Generate array of random variates drawn from self
        :param size: integer or tuple with number of desired sample vectors,
            where each sample.shape == self.shape
        :param rng: (optional) instance of np.random.Generator or subclass
        :return: x = array of samples
            x.shape == (*size, *self.shape)

        Method: Roth (2013) eq. 2.1
        """
        if size is None:
            result_shape = self.shape
        elif np.ndim(size) == 0:
            result_shape = (size,) + self.shape
        else:
            result_shape = size + self.shape
        (eigval, eigvec) = np.linalg.eigh(self.sigma)
        # eigval = 1D array of eigenvalues, eigvec = 2D array with COLUMN eigenvectors
        y = rng.standard_normal(size=result_shape)
        y *= np.sqrt(eigval)
        y = y @ eigvec.T
        # = samples of Gaussian with mean == 0 and cov = inv(self.sigma)
        v = rng.gamma(size=y.shape[:-1], shape=self.df / 2, scale=2 / self.df)
        # = samples of gamma-distributed V with mean == 1., ONE for each vector in y
        return self.loc + y / np.sqrt(v[..., None])

class MultiStudentGivenMean(MultiStudentRV):
    """Variant with mean conditionally depending on an external random variable
    self.mean = self.loc + self.mean_factor * offset
    Needed for predictive model of population mean trait in ir_group.
    """
    def __init__(self, mean_factor=1., **kwargs):
        """
        :param mean_factor: scalar factor to be applied to external offset array
        :param kwargs: arguments to superclass
        """
        super().__init__(**kwargs)
        self.mean_factor = mean_factor

    def rvs(self, mu_rvs=np.asarray(0.), size=None, rng=RNG):
        """Generate array of random variates drawn from self
        :param size: integer or tuple with number of desired samples,
            where each sample.shape == self.shape
        :param rng: (optional) instance of np.random.Generator or subclass
        :param mu_rvs: (optional) array of externally generated random vectors
            such that conditional self.mean = self.loc + self.mean_factor * mu_rvs
        :return: x = array of samples
            x.shape == (*size, *self.shape)
        """
        if size != mu_rvs.shape[:-1] and size is not None:
            logger.warning(f'Changed to size = {mu_rvs.shape[:-1]} like given mu_rvs.')
        size = mu_rvs.shape[:-1]
        x = super().rvs(size=size, rng=rng)
        return x + self.mean_factor * mu_rvs


class StudentRV:
    """Frozen Student distribution of 1D random vector with INDEPENDENT elements,
    i.e., with density function as a product of scalar Student-t densities.
    """
    def __init__(self, df,
                 loc=np.array(0.).reshape((1,)),
                 scale=np.array(1.).reshape((1,)),
                 rng=None):
        """
        :param df: scalar or 1D array-like, degrees of freedom
        :param loc: 1D array or array-like list of location elements
        :param scale: scalar or 1D array or array-like list of scale parameter(s)
            df, loc, and scale must have broadcast-compatible shapes
        :param rng: (optional) random.Generator object
        """
        self.df = np.asarray(df)
        self.loc = np.asarray(loc)
        self.scale = np.asarray(scale)
        if rng is None:
            self.rng = np.random.default_rng()
        else:
            self.rng = rng

    def __repr__(self):
        return (self.__class__.__name__ + '(' +
                '\n\t'.join(f'{k}= {repr(v)},'
                            for (k, v) in self.__dict__.items()) +
                '\n\t)')

    @property
    def size(self):
        return self.loc.size  # len(self.loc)

    @property
    def mean(self):
        if self.df.ndim > 0:
            m = self.loc + 0.  # copy
            m[self.df <= 1.] = np.nan
            return m
        elif self.df > 1.:
            return self.loc
        else:
            return np.full_like(self.loc, np.nan)

    @property
    def var(self):
        """Variance array
        (all correlations == 0.)
        """
        if self.df.ndim > 0:
            v = self.scale ** 2 * self.df / np.maximum(self.df - 2., np.finfo(float).eps)
            v[self.df <= 2.] = np.inf
            v[self.df <= 1.] = np.nan
            return v
        elif self.df > 2:
            return self.scale ** 2 * self.df / (self.df - 2.)
        elif self.df > 1:
            return np.full_like(self.loc, np.inf)
        else:
            return np.full_like(self.loc, np.nan)

    def logpdf(self, x):
        """ln pdf(x | self)
        :param x: array or array-like list of sample vectors
            must be broadcast-compatible with self.loc
        :return: lp = scalar or array of logpdf values
            lp[...] = ln pdf[x[..., :] | self)
            lp.shape == x.shape[:-1]
        Arne Leijon, 2018-07-08, **** checked by comparison to scipy.stats.t
        """
        d = (x - self.loc) / self.scale
        return np.sum(- np.log1p(d**2 / self.df) * (self.df + 1) / 2
                      - np.log(self.scale)
                      + gammaln((self.df + 1) / 2) - gammaln(self.df / 2)
                      - 0.5 * np.log(np.pi * self.df),
                      axis=-1)

    def rvs(self, size=None):
        """Random vectors drawn from self.
        :param size: scalar or tuple with number of sample vectors
        :return: x = array of samples
            x.shape == (*size, self.size)
        """
        if size is None:
            s = self.size
        elif np.isscalar(size):
            s = (size, self.size)
        else:
            s = (*size, self.size)
        # z_sc = scipy_t.rvs(df=self.df, size=s)
        z = self.rng.standard_t(df=self.df, size=s)
        # = standardized samples
        return self.loc + self.scale * z


# ------------------------------------------------- TEST
if __name__ == '__main__':
    from scipy.stats import multivariate_t

    # --------------------------- Test MultiStudentRV
    d = 1
    nx = 1000000
    m = 1. + np.arange(d)  # np.array([1., 2., 3.])  # np.zeros(d)
    s = 1. + np.arange(d)[::-1]  # np.array([3., 2., 1.])
    sigma = RNG.uniform(size=(d, d), low=-0.9, high=0.9)  # correlations
    sigma = (sigma + sigma.T) / 2.
    sigma[range(d), range(d)] = 1.
    sigma *= s[:, None] * s  # covariance
    (eigval, eigvec) = np.linalg.eigh(sigma)  # must be symmetric and positive semi-definite
    print(f'sigma eigval = {eigval}')
    for df in np.array([1.5, 2.5, 3., 10., 30.]):
        st = MultiStudentRV(df=df, loc=m, sigma=sigma)
        print(f'\n*** Testing {st}')
        print(f'st.meam= {st.mean}')
        print(f'st.cov=\n' + np.array_str(st.cov,
                                                 precision=3, suppress_small=True))

        scipy_x = multivariate_t.rvs(size=nx,
                                     df=df,
                                     loc=m,
                                     shape=sigma)
        if d == 1:
            scipy_x = scipy_x.reshape((nx, d))

        print(f'mean(rvs_scipy)= {np.mean(scipy_x, axis=0)}')
        print(f'cov(rvs_scipy)=\n' + np.array_str(np.cov(scipy_x, rowvar=False),
                                                 precision=3, suppress_small=True))

        print(f'MultiStudentRV.rvs() = {st.rvs()}')
        x = st.rvs(size=nx)
        print(f'mean(x)= {np.mean(x, axis=0)}')
        print(f'cov(x)=\n' + np.array_str(np.cov(x, rowvar=False),
                                         precision=3, suppress_small=True))



"""Main template script to analyze ordinal response data using Item Response Theory.
This template should be copied and edited by the user for the desired application.

The model estimates the distribution of one or more latent traits
that determine the probability of response alternatives in a questionnaire,
for one or several populations, represented by groups of respondents.

The results show trait distributions within and between populations,
for an unseen RANDOM INDIVIDUAL in each population, and
for the POPULATION MEAN for each population,
and reveals jointly credible differences between population means.

Populations should preferably be represented by large groups of respondents,
recruited at random from the populations of interest.

Analysis results show inter-individual variance,
correlations between separate traits, if more than one,
the mapping between traits and items,
and estimated response thresholds for all items in all populations,
but individual trait values are not displayed.

The analysis script always includes three main steps:
1: Define and collect item response data from one or more sources,
    and specify a top directory for all results.
2: Learn an OrdinalItemResponseModel instance using the collected response data.
3: Display the main characteristics of the learned model.

*** Version History
* Version 1.0.0:
2023-08-04, no standardization of trait scales in the learned model
2023-07-21, changed signature of ir_model.OrdinalItemResponseModel.initialize

* Version 0.6.0
2023-07-11, cleanup this script template with examples using new ir_source module
2023-07-06, result file format choice moved from ItemResponseDisplaySet.show() -> .save()
2023-07-01, new module ir_source for pandas.read_xxx for input files
2022-09-15, use pandas.read_xxx functions to access input data files.
2020-06-06, first standalone version for ItemResponseCalc package
2020-05-30, template example based on IOI-HA data analysis
"""
# -------- __main__ check to prevent multiprocessor sub-tasks to re-run this script
if __name__ == '__main__':

    import numpy as np
    import pickle
    from pathlib import Path
    import datetime as dt
    import logging
    # from sqlalchemy import create_engine  # to read from SQL database

    import pandas as pd
    pd.options.future.infer_string = True
    # 2026-02-25: tested with Pandas v 2.3: ready for v 3.0

    from ItemResponseCalc import __version__
    from ItemResponseCalc import ir_logging
    from ItemResponseCalc.item_response_data import Questionnaire
    from ItemResponseCalc.item_response_data import ItemResponseDataSet
    from ItemResponseCalc.ir_source import item_response_table, Tables
    # Tables used to join data from two or more sources as one group
    from ItemResponseCalc.ir_model import OrdinalItemResponseModel
    from ItemResponseCalc.ir_display import ItemResponseDisplaySet
    # from ItemResponseCalc.ir_latent import LatentLogistic, LatentNormal

    # -------------------------------- ItemResponseModel parameters:
    max_n_traits = 4  # None
    # = max number of traits to explain total set of item responses
    # = 1, if the test instrument is known to measure a uni-dimensional trait

    n_samples = 50
    # = number of sampled trait vectors for each respondent
    # = number of sampled threshold-defining parameters for each population,

    TEST_DOWNSAMPLE = 1
    # -> no down-sampling -> all records included
    # TEST_DOWNSAMPLE = 10
    # -> smaller data subset only for faster initial test run

    # ----------------------------- STEP 1: Define input sources and result path:

    timestamp_result = True  # Prevent over-writing result files
    # timestamp_result = False  # Repeated runs will save results in same directory

    TOP_PATH = Path.home() / 'Documents/LeijonKTH/heartech/HA-QualityRegistry'
    TOP_PATH = TOP_PATH / 'IRT-IOI-HA'  # or whatever...
    # Change this example to top-dir for everything to be read or written

    ioiha_data_path = TOP_PATH / 'IOI-HA-data'  # or whatever...
    # path to directory containing item response data files

    result_dir = TOP_PATH / 'IOI-HA-results'  # or whatever...
    # path where results will be saved, for several runs if needed
    result_dir = Path('../../../ItemResponseCalc_test') / 'ioiha-result'  # author TEST

    questionnaire_file = ioiha_data_path / 'IOI-HA-English.txt'  # or whatever...
    # text file containing the questions and response alternatives
    # If not available, a Questionnaire object must be created manually
    # as suggested below.

    if timestamp_result:
        timefmt = '{0.year}-{0.month:02}-{0.day:02}-{0.hour:02}{0.minute:02}'
        result_run = timefmt.format(dt.datetime.now())
        # -> name of sub-folder in result_dir
        save_dir = result_dir / result_run
    else:
        save_dir = result_dir
    save_dir.mkdir(parents=True, exist_ok=True)
    # sub-path for results of this particular run

    model_file = 'irt-model.pkl'
    # = name of saved model in save_dir

    log_file = 'run_irt_log.txt'
    # = name of log file saved in save_dir
    ir_logging.setup(save_dir / log_file)
    logger = logging.getLogger(__name__)
    # logger.setLevel(logging.DEBUG)  # only for test

    logger.info('*** Using ItemResponseCalc v.' + __version__)
    if TEST_DOWNSAMPLE > 1:
        logger.info(f'*** Test with down-sampling {TEST_DOWNSAMPLE}')

    def accept_record(r):
        """User-defined inclusion criterion for accepting a response record,
        applied to responses recoded as integer with origin 0,
        and missing responses encoded as -1.
        :param r: list with integer-encoded responses in ONE record from a source
        :return: boolean == True if the record is acceptable
        """
        n_missing = sum(r_i < 0 for r_i in r)
        return n_missing <= 3
    # ----------------------------------------------------------

    # ------------------ 1A: Define questionnaire instrument:
    # create a Questionnaire object to specify
    # number of items and response alternatives

    # Example: load from text file, edit as needed:
    test_instrument = Questionnaire.load(questionnaire_file)

    # if the text file is not available, create test_instrument manually, e.g.:
    # Example: seven items, all with 5 response levels, like the IOI-HA:
    # r_levels = [5, 5, 5, 5, 5, 5, 5]
    # = list with number of response levels, one element for each item
    # test_instrument = Questionnaire(item_response_levels=r_levels)

    # ------------------ 1B: Define sources of data to be included in analysis:
    data_groups = dict()  # space for all included groups

    # Example: Include xlsx file(s) from Hickson, Australia:
    # Edit as needed:
    au1_file = ioiha_data_path / 'Hickson' / 'Short_Eartrak AUS.xlsx'
    au1 = item_response_table(source=au1_file,
                              items={f'Q0{item}': list(range(1, 6))
                                     for item in range(1, 8)},  # as encoded in the xlsx file
                              accept_fcn=accept_record,  # None -> no restriction
                              sample_factor=TEST_DOWNSAMPLE,
                              index_col=0    # with respondent ID, to avoid duplicate records
                              )
    # NOTE: items = a dict with column header and a sequence of allowed ordinal responses,
    #   as encoded in this input source, not necessarily == Questionnaire.responses.
    # If index_col is defined, only the LAST record from each respondent is included.

    logger.info(f'AU-H-10: Using {au1_file}')
    data_groups['AU-H-10'] = au1

    # # Example: second input source: Combine two csv files into one group:
    # nle1_file = ioiha_data_path / 'Kramer' / 'moeder_final - kopieforArne.csv'
    # logger.info(f'NL-02: Using {nle1_file}')
    #
    # nle1 = item_response_table(source=nle1_file,
    #                            items={f'ioi_v{item}': list(range(1, 6))
    #                                   for item in range(1, 8)},     # as encoded in the csv file
    #                            accept_fcn=accept_record,
    #                            sample_factor=TEST_DOWNSAMPLE)   # no respondent ID given
    #
    # nle2_file = ioiha_data_path / 'Kramer' / 'NLSH-Kramer-T2IOI-HA.csv'
    # logger.info(f'NL-16: Using {nle2_file}')
    #
    # kramer2_columns = {'T2IOIHA1': range(5),  # responses coded as 0,.., 4
    #                    'T2IOIHA2': range(1, 6),  # coded as 1,.., 5
    #                    'T2IOIHA3': range(1, 6),
    #                    'T2IOIHA4': range(5),  # coded as 0,.., 4
    #                    'T2IOIHA5': range(1, 6),
    #                    'T2IOIHA6': range(1, 6),
    #                    'T2IOIHA7': range(1, 6)}
    #
    # nle2 = item_response_table(source=nle2_file,
    #                            items=kramer2_columns,
    #                            accept_fcn=accept_record,
    #                            sample_factor=TEST_DOWNSAMPLE)
    # data_groups['NL-02-16'] = Tables(nle1, nle2)

    # SQL example: access data from a database:
    # engine = create_engine("sqlite+pysqlite:///" + str(sql_path / 'testDataBase.db'))
    # ir_sql = item_response_table(source=engine,
    #                              sql='ioiha',  # table name in database
    #                              index_col='EarRowId',
    #                              items={c: [str(i) for i in range(1, 6)]
    #                                     for c in column_names}
    #                              )  # NOTE: string-encoded responses in this DB
    # data_groups['test'] = ir_sql

    # ------------------ 1C: Collect all into ONE DataSet object:

    ids = ItemResponseDataSet(questionnaire=test_instrument,
                              groups=data_groups)
    # NOT YET reading any response data
    # In case of any input error, it will show here:
    logger.info('Item Response Data = \n' + repr(ids))

    # (Optional) Show response counts, to verify that input was OK
    r_counts = ids.response_counts()
    # = list with one TableRef object for each questionnaire item
    logger.info('Response Counts =\n'
                 + '\n\n'.join(item.question + '\n' + c_df.to_string()
                               for (item, c_df) in zip(ids.questionnaire.items,
                                                       r_counts)
                               )
                 )
    # ----- Optionally, save item count tables in separate files, too
    # item_counts = save_dir / 'data_counts'
    # item_counts.mkdir(parents=True, exist_ok=True)
    # for item_i in r_counts:
    #     item_i.save(item_counts, table_format='csv')    # for input to other package

    # ------------------ STEP 2: Learn ItemResponseModel from data:

    irm = OrdinalItemResponseModel.initialize(ids,
                                              # latent_class=LatentLogistic,    # default
                                              # n_traits=max_n_traits,  # default = n_items
                                              n_samples=n_samples,
                                              # restrict_traits=True,  # default
                                              # restrict_threshold=False,  # default
                                              # threshold_pseudo_factor=1.,   # default
                                              # trait_scale=3.,  # default
                                              )
    # restrict_traits = True -> Global mean trait fixed at zero.
    # restrict_threshold = True -> Middle response threshold fixed at zero for all items.
    # NOTE: Even with restrict_threshold = False,
    #   response thresholds must be rather similar across populations,
    #   although some variations are allowed. Therefore,
    #   the model includes a number of "pseudo_groups" = threshold_pseudo_factor * real n_groups,
    #   all with ZERO threshold deviations from the global mean.
    #   Larger threshold_pseudo_factor -> smaller threshold variations between populations.

    logger.info(f'Learning model with max {irm.n_traits} traits'
                 + f' for {irm.n_groups} groups'
                 + f' with {irm.n_subjects} included respondent records in total')
    logger.info(f'The model uses {n_samples} samples for each respondent group')

    LL = irm.learn(max_hours=0, max_minutes=30)
    # *** Might take hours, trying with shorter time limit first.
    # When learning is complete, it will finish before the time limit.

    logger.info(f'Learned model with {irm.n_groups} groups'
                 + f' with {irm.n_subjects} subjects in total')
    # logger.info('Log Likelihood: ' + np.array2string(np.array(LL), precision=1))

    # ------------------ Optionally, save model for later display generation
    with (save_dir / model_file).open('wb') as f:
        pickle.dump(irm, f)
    logger.info(f'Model saved in {save_dir}')
    logger.info('Learned model = \n' + repr(irm))

    irm.prune()  # keep only traits that were really needed to model the data

    # ------------------ STEP 3: Display Model Results:
    # mapping_item used to display trait scales by reference to ONE item
    ir_display = ItemResponseDisplaySet.show(irm,
                                             mapping_item=-1,  # last questionnaire item
                                             percentiles=[2.5, 50., 97.5],
                                             mpl_params={'figure.max_open_warning': 0,  # suppress warning
                                                         'figure.autolayout': True,  # -> tight layout
                                                         'axes.labelsize': 'x-large'},  # -> matplotlib.rcParam
                                             # mpl_style='my_style_sheet',
                                             # ... any other ema_display.FMT settings
                                             # ... any ema_display_format.FMT settings
                                             )
    ir_display.save(save_dir,
                    figure_format = 'pdf',  # any format that Matplotlib can handle
                    table_format = 'txt',   # any format that Pandas can handle
                    float_format = '%.2f')  # any other parameter(s) for pandas.write_xxx

    # --------------------------- (optionally) save in other format(s), too:
    # ir_display.save(save_dir,
    #          table_format='csv',  # only tables, for input to other package
    #          float_format='%.4f',
    #          )
    # ir_display.save(save_dir,
    #          figure_format='eps',  # only figures, for input to other package
    #          )

    logger.info(f'All results saved in {save_dir}')

    # ------------------------------ Instrument Reliability Measures:
    mi = irm.instrument_mutual_information()  # bits
    person_separation = 2.**mi
    logger.info(f'Instrument-trait information = {mi:.2f} bits. '
                + f'Separation ability: {person_separation:.1f} person categories.')
    rho = 1. - 2**(-2 * mi)
    logger.info(f'Equivalent Reliability Coefficient= {rho:.3f}')

    # ------------------ (Optional) PCA of inter-individual trait covariance within populations
    if irm.n_traits > 1:
        cov_within = irm.predictive_individual_cov()
        (eig_val, eig_vec) = np.linalg.eigh(cov_within)
        logger.debug('Cov eig_val= ' +
                      np.array2string(eig_val, precision=3))
        logger.debug('Cov eig_vec=\n' +
                      np.array2string(eig_vec, precision=3))
        eig_val_cum = np.cumsum(eig_val[::-1]) / np.sum(eig_val)
        logger.info(f'One, Two, etc Principal Trait Factors: ' +
                     ', '.join(f'{ev:.1%}'
                               for ev in eig_val_cum) +
                     ', of Variance within Populations')

    logging.shutdown()



"""Package ItemResponseCalc implements Bayesian analysis of
subjective responses to items in a questionnaire,
using Item Response Theory (IRT) applied with
the Graded Response Model (Samejima, 1997; Fox, 2010).

This package is only intended for the analysis of data with ORDINAL responses,
e.g., the IOI-HA questionnaire for users' satisfaction with their hearing aids.

The IRT Graded Response Model treats subjects' responses
as determined by the outcome of a latent individual trait variable,
i.e., similar to the latent internal "sensation"
variable assumed to determine responses in psycho-physical experiments.
This package uses a logistic distribution for the latent variables.

Each questionnaire item corresponds to exactly ONE individual trait variable,
but each trait variable may determine the subject's responses to several items.

Therefore, for a questionnaire with I separate items,
the model can learn D <= I trait variables, based on the observed responses.
The model automatically learns how many
different trait variables are needed to 'explain' the recorded data.

NOTE: In most modules of this package,
the ordinal response to the i-th item is encoded as
an integer in {0, 1,..., L_i-1},
where L_i = the number of ordinal response levels for the i-th item.
Any value outside this range is interpreted as a missing response.

However, in data files with recorded responses,
the recorded responses are sometimes encoded as integers in {1, 2,..., L_i},
with missing values encoded as zero.
Other response files may instead store the actual text labels
on the response alternatives in the questionnaire.

Therefore, the package user must specify the response encoding
separately for each input data source.

*** Usage

Copy the template script `run_irt.py` to your working directory, rename it, and
edit the copy as suggested in the template, to specify
    - your experimental layout,
    - your input data file(s),
    - a directory where all output result files will be stored.

*** Main Modules

ir_base:
    General help methods to access sampled data

item_response_data:
    Interface to one or several data sets containing individual responses
    from respondents recruited from one or more separate population(s).
    See this module for information about input files and data formats.

ir_model:
    The core of the Bayesian IRT model adapted to all observed data.

ir_global:
    global parameter distributions, common for all included populations.

ir_graded_response_prob:
    Calculates response probabilities with the Graded Response Model

ir_group:
    Classes representing individual- and population-specific models
    defining trait distributions for each individual in ONE group of respondents,
    and response thresholds for the population represented by the group.

ir_display:
    Functions and classes to display all analysis results as figures and tables.
    
ir_source:
    Implements access methods for input data in several source types,
    e.g., Excel or CSV files, or SQL database,
    using reader functions available in package Pandas,
    or a similar user-defined reader function.

ir_thresholds:
    Mapping response thresholds from internal model parameters

Analysis results are saved as figures and tables in the selected result directory.
After running an analysis, the logging output briefly explains
the analysis results presented in figures and tables.

*** References
* For Version >=1.0:
A. Leijon: Analysis of Ordinal Response Data using Bayesian Item Response Theory package ItemResponseCalc.
    Documentation report with all math details for the new model structure

* For Version <= 0.6:
A. Leijon, H. Dillon, L. Hickson, M. Kinkel, S. E. Kramer, and P. Nordqvist.
    Analysis of data from the international outcome inventory for hearing aids (IOI-HA)
    using Bayesian item response theory. Int J Audiol 60(2):81–88, 2020.
    doi: 10.1080/14992027.2020.1813338
J.-P. Fox (2010). *Bayesian Item Response Modeling: Theory and Applications*.
    Statistics for Social and Behavioral Sciences. Springer.
G. N. Masters (1982). A Rasch model for partial credit scoring.
    *Psychometrika*, 47(2):149–174.
F. Samejima (1997). Graded response model.
    In W. J. v. D. Linden and R. K. Hambleton, eds.,
    *Handbook of Modern Item Response Theory*, p. 85–100. Springer, New York.


*** Version History
* Version 1.0.1-2: Minor bug fix. Tested with Pandas v 2.3. Should be OK with v 3.0.

* Version 1.0.0: New model structure. Item response thresholds are separate for each group.
    Thus, differing response patterns between included groups can be "explained"
    partly by their different interpretations of response categories,
    rather than by real perceptual differences between populations.
    This is a reasonable assumption, in particular, if respondents from different populations
    have used different language versions of the questionnaire items.

* Version 0.6.0:
2023-07-07, using package pandas to access input data files, and for tabulated results.

* Version 0.5:
2020-06-xx, prepare to allow different item response probability models-
2020-06-08, first version uploaded to PyPi, tested with several IOI-HA data sets.
2019-08-24, first functional version with automatic selection of trait dimensionality.
"""

__name__ = 'ItemResponseCalc'
__version__ = '1.0.2'
__all__ = ['__version__', 'run_irt']





"""Help classes for distribution of latent sensory variable
used in an ItemResponseCalc model.

*** Version history:
* Version 1.0.0: new module for this version, allow either logit or probit model.
2023-08-30, LatentNormal methods -> staticmethod
2023-07-20, module copied from EmaCalc
"""
# *** Calculations for d_log_cdf_diff(tau_a - theta, tau_b - theta)
# *** separately w.r.t tau and theta ?
# In the Logistic model, the derivative w.r.t theta is independent of interval width tau_b - tau_a

import numpy as np
from scipy.special import expit, log_expit, log_ndtr
import logging

logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)  # *** TEST


# ------------------------------------------------------------------
class LatentLogistic:
    """Distribution of latent decision variable = standard logistic
    cdf(x) = 1 / (1 + exp(-x)) = expit(x)
    """
    unit_label = 'logit unit'
    # for axis label in attribute plots
    scale = np.pi / np.sqrt(3.)
    # = st.dev. of standard logistic distribution
    # may be used to standardize result scale in displays

    @staticmethod
    def log_cdf_diff(a, b):
        """log prob( a < Z <= b)
        where Z is a standard logistic random variable
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( -inf <= a < b <= +inf)
        :return: log_p = array with log probabilities, element-wise
            log_p.shape == a.shape == b.shape
        """
        # p = expit(b) - expit(a)
        # *** may be inaccurately == 0. if 36.7. < a < b in double prec.
        # *** problem when b = inf and a > 36.7  -> d == 0.
        # *** No problem with b = -100. and a = -inf

        d = b - a
        zero_d = d == 0.
        if np.any(zero_d):  # *** precaution to detect numerical underflow
            logger.warning(f'LatentLogistic.log_cdf_diff: {np.sum(zero_d)} non-positive interval width. '
                           + 'Should never happen! Contact the author.')
            logger.warning(f'a =\n' + np.array_str(a[zero_d]))
            logger.warning(f'b =\n' + np.array_str(b[zero_d]))
            logger.warning(f'd = b-a =\n' + np.array_str(d[zero_d]))
            d[zero_d] = np.finfo(float).tiny  # to avoid expm1(0.) -> 1 / 0.
        # *** Safe method for -inf < b <= +inf;  -inf <= a < +inf; 0 < d == b-a <= +inf:
        # log(expit(b) - expit(a)) = log_expit(b) + log_expit(-a) + log[1. - exp(-(b-a))]
        lp = log_expit(b) + log_expit(-a)
        log_1mexp = np.log1p(-np.exp(-d))
        # but log[1. - exp(-(b-a))] may still get truncated -> 0. for very small d = b-a
        small_d = d < 0.001
        log_1mexp[small_d] = np.log(np.expm1(d[small_d])) - d[small_d]
        lp += log_1mexp
        return lp

    @staticmethod
    def d_log_cdf_diff(a, b):
        """Element-wise partial derivatives of log_cdf_diff(a, b)
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: tuple (dll_da, dll_db) of arrays, where
            dll_da[...] = d log_cdf_diff(a[...], b[...]) / d a[...]
            dll_db[...] = d log_cdf_diff(a[...], b[...]) / d b[...]
            dll_da.shape == dll_db.shape == a.shape == b.shape
        2023-06-08, tested by finite-diff comparison
        """
        d = b - a
        zero_d = d == 0.
        if np.any(zero_d):  # *** precaution to detect numerical underflow
            logger.warning(f'LatentLogistic.log_cdf_diff: {np.sum(zero_d)} non-positive interval width. '
                           + 'Should never happen! Please contact the author.')
            d[zero_d] = np.finfo(float).tiny  # to avoid expm1(0.) -> 0.
            # No problem if this tiny logprob multiplied by zero count later
        # log_cdf_diff(a,b) = log_expit(b) + log_expit(-a) + log[1. - exp(-(b-a))]
        # d log_expit(-a) / d a = - expit(a)
        # d log_expit(b) / d b = expit(-b)
        # d log[1. - exp(-(b - a))] / d a = - 1. / expm1(d)
        # d log[1. - exp(-(b - a))] / d b =   1. / expm1(d)
        d_log_1mexp = 1. / np.expm1(d)
        return - expit(a) - d_log_1mexp, expit(-b) + d_log_1mexp


class LatentNormal:
    """Distribution of decision variable = standard normal,
    i.e., Thurstone Case V,
    with distribution function cdf(x) = Phi(x)
    """
    unit_label = 'd-prime unit'
    scale = 1.

    # @classmethod  # ****** -> staticmethod
    @staticmethod
    def log_cdf_diff(a, b):
        """log[ P( a < Z <= b) ], where Z is a Gaussian standard random variable.
        Numerically stable, avoiding numerical under- or over-flow for -inf <= a < b <= +inf.
        OK also for very small interval width (b - a).
        :param a: = array with LOWER interval limits
        :param b: = array with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: log_p = log(ndtr(b) - ndtr(a)) = log(ndtr(-a) - ndtr(-b))
            log_p.shape == a.shape == b.shape
        """
        low = np.minimum(a, -b)
        high = np.minimum(b, -a)
        # interval (low, high) = either (a, b) or (-b, -a)
        # same in theory, but lower variant is numerically more accurate
        ln_ndtr_low = log_ndtr(low)
        ln_ndtr_high= log_ndtr(high)
        zero_d = ln_ndtr_low >= ln_ndtr_high
        if np.any(zero_d):  # *** precaution to detect numerical underflow
            logger.warning(f'LatentNormal.log_cdf_diff: {np.sum(zero_d)} non-positive probabilities. '
                           + 'Should never happen! Contact the author.')
            logger.warning(f'a =\n' + np.array_str(a[zero_d]))
            logger.warning(f'b =\n' + np.array_str(b[zero_d]))
            logger.warning(f'd = b-a =\n' + np.array_str((b-a)[zero_d]))
        return ln_ndtr_high + np.log1p(-np.exp(ln_ndtr_low - ln_ndtr_high))

    # @classmethod
    @staticmethod
    def d_log_cdf_diff(a, b):
        """Element-wise partial derivatives of log_cdf_diff(a, b)
        :param a: = array with LOWER interval limits
        :param b: = arrays with UPPER interval limits
            a.shape == b.shape
            all( a < b )
        :return: tuple (dll_da, dll_db), where
            dll_da[...] = d log_cdf_diff[a[...], b[...]) / d a[...]
            dll_db[...] = d log_cdf_diff[a[...], b[...]) / d b[...]
            dll_da.shape == dll_db.shape == a.shape == b.shape
        """
        # cdf_diff(a, b) = ndtr(b) - ndtr(a)
        # d log(cdf_diff) / d a = - pdf(a) / cdf_diff(a, b)
        # d log(cdf_diff) / d b =   pdf(b) / cdf_diff(a, b),
        # where pdf(x) = pdf of standard normal (Gaussian)
        ln_cdf_diff = cls.log_cdf_diff(a, b)
        return - np.exp(_log_npdf(a) - ln_cdf_diff),  np.exp(_log_npdf(b) - ln_cdf_diff)


# ------------------------------------------------- module local

_log_sqrt_2pi = np.log(2 * np.pi) / 2
# = module constant


def _log_npdf(x):
    """log pdf of standard normal distribution
    :param x: scalar or array
    :return: lp = log pdf(x)
        lp.shape == x.shape
    """
    return - x**2 / 2 - _log_sqrt_2pi


# ------------------------------------------------- TEST:
if __name__ == '__main__':
    from scipy.optimize import approx_fprime, check_grad

    print('*** Testing LatentLogistic and LatentNormal derivatives')

    # --------------------------------------------------
    for cls in (LatentLogistic, LatentNormal):

        print(f'\nTesting {cls}.d_log_cdf_diff() with 1D (a, b) args')
        tau = np.array([-np.inf, 0., 1e-5, np.inf])  # 1.e-7 also OK with LatentLogistic
        # tau = np.array([-50.1, -50.])
        # tau = np.array([50., 50.1])
        th = -0.  # 5000, -5000 also OK
        (a, b) = (tau[:-1] - th, tau[1:] - th)

        print(f'cls.log_cdf_diff({a}, {b}) = \n', cls.log_cdf_diff(a, b))

        for i in range(len(a)):
            ind = [i]
            def fun_a(a):
                return cls.log_cdf_diff(a, b[ind])

            def jac_a(a):
                return cls.d_log_cdf_diff(a, b[ind])[0]


            def fun_b(b):
                return cls.log_cdf_diff(a[ind], b)

            def jac_b(b):
                return cls.d_log_cdf_diff(a[ind], b)[1]

            print('')
            print(f'approx d_log_cdf_diff_da[{ind}] = ', approx_fprime(a[ind], fun_a, epsilon=1e-8))
            print(f'exact  d_log_cdf_diff_da[{ind}] = ', jac_a(a[ind]))
            err = check_grad(fun_a, jac_a, a[ind], epsilon=1e-8)
            print('check_grad err = ', err)

            print(f'approx d_log_cdf_diff_db[{ind}] = ', approx_fprime(b[ind], fun_b, epsilon=1e-8))
            print(f'exact  d_log_cdf_diff_db[{ind}] = ', jac_b(b[ind]))
            err = check_grad(fun_b, jac_b, b[ind], epsilon=1e-8)
            print('check_grad err = ', err)



"""Functions and classes to display results from an OrdinalItemResponseModel object.
** Version history:

* Version 1.0.0:
2023-08-12, minor bugfix

* Version 0.6:
2023-07-07, rely mainly on matplotlib.rcParams for plot styles
2023-07-06, define name of FigureRef and ResultTable already when created
2023-07-05, new table result functions using Pandas for result tables

* Version 0.5: first published version
"""
import numpy as np
from itertools import cycle, product

import matplotlib.pyplot as plt
import logging

import pandas as pd

logger = logging.getLogger(__name__)

# --------------------------- Default format parameters:
FMT = {'colors': 'rbg',  # color cycle to separate results in plots
       'line_styles': ['solid', 'dashed', 'dashdot', 'dotted'],
       'markers': 'o^sDv',  # plot symbols, each repeated with each fillstyle
       'fillstyle': ['full', 'none'], # marker style
       'sum': 'Sum',
       }
# = module-global dict with default settings for display details
# that may be changed by user

# NOTE: FMT['colors'], FMT['linestyles'], FMT['markers'], and FMT['fillstyle']
#   override any corresponding properties in matplotlib.rcParams,
#   because rcParams.axes.prop_cycle allows only equal lengths of these property sequences.
#   Here, the UNEQUAL lengths of FMT default sequences will combine
#   into a sequence with many combinations, before the style repeats itself.


class FileWriteError(RuntimeError):
    """Any type of exception while attempting to write a pd.DataFrame table to a file
    """


def set_format_param(mpl_style=None, mpl_params=None, **kwargs):
    """Set / modify format parameters.
    Called before any displays are generated.
    :param mpl_style: (optional) matplotlib style sheet, or list of style sheets
    :param mpl_params: (optional) dict with matplotlib (k, v) rcParam settings
    :param kwargs: dict with any formatting variables to be stored in FMT
    :return: None
    """
    if mpl_style is not None:
        plt.style.use(mpl_style)
    if mpl_params is not None:
        plt.rcParams.update(mpl_params)
    other_fmt = dict()
    for (k, v) in kwargs.items():
        k = k.lower()
        if k in FMT:
            FMT[k] = v
        else:
            other_fmt[k] = v
    if len(other_fmt) > 0:
        logger.warning(f'Parameters {other_fmt} unknown, not used.')


# ---------------------------- Main Result Classes
class FigureRef:
    """Reference to a single graph instance
    """
    def __init__(self, ax, name=None):
        """
        :param ax: Axes instance containing the graph
        :param name: string name of figure
        """
        self.ax = ax
        self.name = name

    @property
    def fig(self):
        return self.ax.figure

    def save(self, path,
             figure_format):
        """Save figure to given path
        :param path: Path to directory where figure is saved
        :param figure_format: file format of saved figure
        :return: None
        """
        if figure_format is not None:
            name = _clean_file_name(self.name)
            path.mkdir(parents=True, exist_ok=True)
            f = (path / name).with_suffix('.' + figure_format)
            try:
                self.fig.savefig(f)
            except Exception as e:  # any error, just warn and continue
                logger.warning(f'Could not save plot to {f}. Error: {e}')

    def mapped_y_axis(self, y, y_mapped, y_label=''):
        """Add a second y_axis to self.ax with ticks placed at
        y values corresponding to uniform steps in y_mapped values
        but labelled by the y_mapped values
        :param y: long sequence of y_values along original y axis
        :param y_mapped: corresponding transformed y_values for new y_ticks
            len(y_mapped) == len(y)
            y and y_mapped monotonically increasing
        :param y_label: label of second y-axis
        :return: None
        """
        ax2 = self.ax.twinx()
        ax2.set_ylim(self.ax.get_ylim())
        ymap_ticks = _nice_ticks(np.amin(y_mapped), np.amax(y_mapped))
        # = uniform ticks in y_mapped scale
        ymap_ticklabels = [f'{tick:.1f}' for tick in ymap_ticks]
        y_ticks = np.interp(ymap_ticks, y_mapped, y)
        # = transformed to corresponding y scale positions
        ax2.set_yticks(y_ticks)
        ax2.set_yticklabels(ymap_ticklabels)
        ax2.set_ylabel(y_label)
        # self.fig.tight_layout()


# --------------------------------------------------
class Table(pd.DataFrame):
    """Subclass adding a general save method,
    automatically switching to desired file format.
    """
    def save(self, file_path,
             allow_over_write=True,
             write_fcn=None,
             **file_kwargs):
        """Save self to a table-style file.
        :param file_path: Path instance defining file location and full name incl. suffix.
        :param allow_over_write: (optional) boolean, if False, find new unused path stem.
        :param write_fcn: (optional) user-supplied function with signature
            write_fcn(table, path, **kwargs).
            If None, default pd.DataFrame method is used, determined by file_path.suffix
        :param file_kwargs: (optional) any additional arguments for the
            write function for the specific file format.
        :return: None
        """
        if not allow_over_write:
            file_path = safe_file_path(file_path)
        suffix = file_path.suffix
        try:
            # ******** file_path.parent.mkdir here? NO, done by container class
            if write_fcn is None:
                if suffix in ['.xlsx', '.xls', '.odf', '.ods', '.odt']:
                    self.to_excel(file_path, **file_kwargs)
                elif suffix in ['.csv']:
                    self.to_csv(file_path, **file_kwargs)
                elif suffix in ['.txt']:
                    self.to_string(file_path, **file_kwargs)
                elif suffix in ['.tex']:
                    # with warnings.catch_warnings():
                    #     # suppress Pandas FutureWarning about to_latex method
                    #     warnings.simplefilter('ignore')
                    #     self.to_latex(file_path, **file_kwargs)
                    self.to_latex(file_path, **file_kwargs)
                else:
                    raise FileWriteError(f'No DataFrame write method for file type {suffix}')
            else:
                write_fcn(self, file_path, **file_kwargs)
        except Exception as e:
            raise FileWriteError(f'Could not write to {file_path}. Error: {e}')

class ResultTable(Table):
    """A pd.DataFrame table subclass, with a name and special save method
    """
    def __init__(self, df, name):
        """
        :param df: a Table(pd.DataFrame) instance
        :param name: file name for saving the table
        """
        super().__init__(df)
        self.name = name

    def save(self, path,
             table_format=None,
             **kwargs):
        """Save table to file.
        :param path: Path to directory for saving self.
        :param table_format: table-format string code -> file-name suffix
        :param kwargs: (optional) any additional arguments to pandas writer function
        :return: None
        """
        if table_format is not None:
            name = _clean_file_name(self.name)
            f = (path / name).with_suffix('.' + table_format)
            try:
                super().save(f, **kwargs)
            except Exception as e:  # any error, just warn and continue
                logger.warning(f'Could not save result table. Error: {e}')


class DiffTable(ResultTable):
    """Special subclass suppressing index in save method
    """
    def save(self, path,
             **kwargs):
        """Save table to file.
        :param path: Path to directory for saving self.
        :param kwargs: (optional) any additional arguments to pandas writer function
        :return: None
        """
        if 'index' not in kwargs:
            kwargs['index'] = False  # override Pandas default = True
        super().save(path, **kwargs)


# ------------------------------------------------ plot routines

def fig_log_likelihood(learned_ll, n_users,
                       name,
                       title='',
                       **kwargs):
    """plot VI learning result
    :param learned_ll = list log-likelihood values from a learned ItemResponseModel
    :param n_users = scalar number of users included in the total log-likelihood
    :param name = figure name for file
    :param title = (optional) figure title
    :param kwargs = (optional) dict with any additional arguments for plot function
    :return: None
    """
    f_logprob, ax = plt.subplots()
    t1 = 1
    # for ll in learned_ll:
    t = [t1 + n for n in range(len(learned_ll))]
    # n_train = ham.n_training_users
    ax.plot(t, np.array(learned_ll) / n_users, **kwargs)
    # t1 = t[-1] + 1
    ax.set_xlim(0, t[-1]+1)
    ax.set_xlabel('Learning Iterations')
    ax.set_ylabel('Log-likelihood / N')
    if 0 < len(title):
        ax.set_title(title)
    return FigureRef(ax, name=name)


def fig_response_freq(count, name):
    """Generate a plot with relative frequencies of response counts for all items
    :param count: list of Counter objects  *** dict ?
        count[i][0] = number of missing responses for i-th item
        count[i][l] = number of responses in l-th ordinal level for i-th item
    :param name = string name of this figure
    :return: FigureRef object
    """
    fig, ax = plt.subplots()
    for (i, (c, col, (m, fill_style))) in enumerate(zip(count,  # total_count,
                                                        cycle(FMT['colors']),
                                                        cycle(product(FMT['markers'],
                                                                      FMT['fillstyle'])))):
        x_i = np.arange(1 + max(*c.keys()))  # key = 0 means missing data
        y_i = np.array([c[r] for r in x_i])
        y_i = y_i / np.sum(y_i)
        y_i *= 100  # in percent
        ax.plot(x_i, y_i, label = f'Q{i+1}', color=col,
                marker=m, fillstyle=fill_style)
    ax.set_xticks(np.arange(0, 6))
    ax.set_xticklabels(['Missing'] + [f'R={i+1}' for i in range(5)])
    ax.set_xlabel('Item Response')
    ax.set_ylabel('Rel. Frequency (%)')
    ax.legend(loc='best')
    return FigureRef(ax, name=name)


def fig_response_prob(t, p,
                      name,
                      tau=None,
                      x_label='',
                      y_label=''):
    """figure with response prob vs. trait for ONE item
    :param t: 1D array with trait values
    :param p: 2D array with response probabilities
        p[n, l] = P{l-th response | t[n]}
    :param name: string for plot file name
    :param tau: (optional) 1D array with trait thresholds
    :param x_label: string for x-axis label
    :param y_label: string for y-axis label
    :return: FigureRef object
    """
    fig, ax = plt.subplots()
    for (l, (p_l, c, ls)) in enumerate(zip(p.T,
                                           cycle(FMT['colors']),
                                           cycle(FMT['line_styles']))):
        ax.plot(t, p_l, color=c, linestyle=ls, label=f'R={1+l}')
    if tau is not None:
        # plot ticks at thresholds
        x = [tau, tau]
        y = [np.zeros(len(tau)), 0.1 * np.ones(len(tau))]
        ax.plot(x, y, color='k', linewidth=1., linestyle='solid')
    ax.set_ylim(0., 1.)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.legend(loc='best')
    return FigureRef(ax, name=name)


def fig_fisher_information(t, f, item_labels,
                           name,
                           x_label='',
                           y_label=''):
    """figure with response prob vs. trait for ONE item
    :param t: 1D array with hypothetical TRUE trait values
    :param f: 2D array with Fisher Information data
        f[i, n] = Fisher Information for i-th item given true trait=t[n]
    :param item_labels: list with string labels
        len(item_labels) == f.shape[0]
    :param name: string for plot file name
    :param tau: (optional) 1D array with trait thresholds
    :param x_label: string for x-axis label
    :param y_label: string for y-axis label
    :return: FigureRef object
    """
    fig, ax = plt.subplots()
    for (f_i, id_i, c, ls) in zip(f,
                                  item_labels,
                                  cycle(FMT['colors']),
                                  cycle(FMT['line_styles'])):
        ax.plot(t, f_i, color=c, linestyle=ls, label=id_i)
    if f.shape[0] > 1:  # plot sum, too
        ax.plot(t, np.sum(f, axis=0), color='k', linestyle='solid',
                linewidth=1.3*plt.rcParams['lines.linewidth'],
                label=FMT['sum'])
    # ax.set_ylim(0., 1.)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    ax.legend(loc='best')
    return FigureRef(ax, name=name)


def fig_thresholds(tau_group,
                   name,
                   x_label='Population',
                   y_label='Thresholds'):
    """figure with threshold quartiles for ONE item and all groups
    :param tau_group: pd.DataFrame with one column for each group,
        and one MultiIndex row for each (Response, Quartile),
        with UPPER threshold for each ordinal response.
        Currently exactly THREE quantiles: first quartile, median, third quartile
    :param name: string label for the plot file
    :param x_label: (optional) string for x label
    :param y_label: (optional) string for y label
    :return: FigureRef instance
    """
    if tau_group is None:
        return None
    (fig, ax) = plt.subplots()
    responses = tau_group.index.levels[0]
    q = tau_group.index.levels[1]
    for r in responses:  # each response
        t = tau_group.loc[r].to_numpy()
        x = np.arange(t.shape[-1])
        ax.plot(np.tile(x, (2, 1)), t[[0, -1], :], 'r-')
        ax.plot(x[:, None], t[q==0.5, :].T, 'r-', marker='o', linewidth=0.5)
    ax.set_xticks(np.arange(tau_group.shape[-1]))
    ax.set_xticklabels(tau_group.columns.values)
    ax.set_xlabel(x_label)
    ax.set_ylabel(y_label)
    return FigureRef(ax, name=name)



def fig_percentiles(perc,
                    y_label,
                    x_ticklabels,
                    name,
                    case_labels=None,
                    x_label=None,
                    x_offset=0.1,
                    x_space=0.5,
                    **kwargs):
    """create a figure with trait percentile results
    :param perc: 3D (or 2D) array with trait percentiles
        perc[c, p, t] = p-th percentile of t-th variable for c-th case
        percentiles plotted vertically vs t-variables horizontally,
    :param y_label: string for y axis label
    :param x_label: string for x-axis label
    :param x_ticklabels: list of strings with labels for x_ticks,
        one for each value in rows perc[..., :]
        len(x_ticklabels) == perc.shape[-1] == number of traits
    :param name: string name of this figure
    :param case_labels: list of strings with labels for cases, if more than one
        len(case_labels) == perc.shape[0] if perc.ndim == 3
    :param x_offset: (optional) horizontal space between case plots for each x_tick
    :param x_space: (optional) min space outside min and max x_tick values
    :param kwargs: (optional) dict with any additional keyword arguments for plot commands.
    :return: None  # fig object with single plot axis with all results
    """
    fig, ax = plt.subplots()
    if perc.ndim == 2:
        perc = perc[np.newaxis,...]
        case_labels = [None]
    (n_cases, n_perc, n_xticks) = perc.shape
    x = np.arange(0., n_xticks) - x_offset * (n_cases - 1) / 2
    for (c_key, c_y, c, (m, fs)) in zip(case_labels, perc,
                                      cycle(FMT['colors']),
                                      cycle(product(FMT['markers'],
                                                    FMT['fillstyle']))):
        ax.plot(np.tile(x, (2, 1)), c_y[[0, 2], :],
                linestyle='solid', color=c, **kwargs)
        ax.plot(x, c_y[1, :], linestyle='', color=c,
                marker=m, fillstyle=fs,  # markeredgecolor=c,  # markerfacecolor='w',
                label=c_key, **kwargs)
        x += x_offset
    (x_min, x_max) = ax.get_xlim()
    x_min = min(x_min, -x_space)
    x_max = max(x_max, n_xticks - 1 + x_space)
    ax.set_xlim(x_min, x_max)
    ax.set_xticks(np.arange(n_xticks))
    ax.set_xticklabels(x_ticklabels)
    ax.set_ylabel(y_label)  # + ' (' + y_unit + ')')
    ax.set_xlabel(x_label)
    if n_cases > 1:
        # make space for legend
        (x_min, x_max) = ax.get_xlim()
        ax.set_xlim(x_min, x_max + 0.6)
        ax.legend(loc='best')
    if name is None:
        name = _clean_file_name(y_label)
    return FigureRef(ax, name=name)


def fig_credible_diff(c_diff,
                      diff_label,
                      x_labels,
                      x_label=None,
                      title=None,
                      cred_levels=None):
    """Nice color plot to show jointly credible differences
    :param c_diff: list of tuples ((i, j), p), indicating that
        x_labels[i] > x-labels[j] with joint credibility p
    :param diff_label: name for this display figure
    :param x_labels: list of labels of compared categories
    :param x_label: (optional) axis label
    :param title: (optional) string with plot title
    :param cred_levels: (optional) list of joint-credibility values, in DECREASING order
    :return:
    """
    if cred_levels is None:
        cred_levels = [.99, .95, .9, .8, .7]
    marker_sizes = [15, 12, 9, 6, 3]

    cred_levels = np.asarray(cred_levels)
    symbol_size = None  # use previous size to detect change

    # ------------------------------------------------
    def select_symbol(p, prev_size):
        """Determine symbol size and legend text
        """
        i_size = np.nonzero(cred_levels < p)[0][0]
        size = marker_sizes[i_size]
        if size != prev_size:
            return size, f'> {cred_levels[i_size]:.0%}'
        else:
            return size, None
    # ------------------------------------------------

    fig, ax = plt.subplots()
    for ((i, j), p) in c_diff:
        symbol_size, label = select_symbol(p, symbol_size)
        if symbol_size > 0:
            s, = ax.plot(i, len(x_labels) - 1 - j, 'sr', markersize=symbol_size)
            if label is not None:
                s.set_label(label)
    ax.set_xlim([-0.5, len(x_labels) - 0.5])
    ax.set_ylim([-0.5, len(x_labels) - 0.5])
    ax.set_xticks(np.arange(len(x_labels)))
    ax.set_yticks(np.arange(len(x_labels)))
    ax.set_xticklabels(x_labels,
                       **_x_tick_style(x_labels)
                       )
    ax.set_yticklabels(x_labels[::-1],  # reversed order
                       rotation='horizontal',
                       horizontalalignment='right'
                       )
    if x_label is not None:
        ax.set_xlabel(x_label)
        ax.set_ylabel(x_label)
    ax.set_title(title)
    if len(c_diff) > 0:
        ax.legend(loc='best')
    fig.tight_layout()
    return FigureRef(ax, name=diff_label)


# --------------------------------------- Table formatting functions
def tab_correlation_matrix(cov, item_trait_map, name='Corr', header='Trait'):
    """Normalized correlation matrix between IRT traits corresponding to items.
    :param cov: estimated, possibly un-normalized, covariance matrix,
        possibly including traits that have NO item correspondence
    :param item_trait_map: 2D boolean array mapping trait to item
        item_trait_map[i, t] == True <=> t-th trait determines i-th item response
    :param name: (optional) string name of this table
    :param header: (optional) name for row and column index
    :return: a ResultTable instance
    """
    trait_sel = np.any(item_trait_map, axis=0)
    c = cov[:, trait_sel][trait_sel, :]
    std = np.sqrt(np.diag(c))
    # Normalize covariance matrix
    c /= std
    c /= std[:, None]
    trait_labels = make_trait_labels(item_trait_map)
    # *** use externally defined trait labels? ******
    trait_index = pd.Index(trait_labels, name=header)
    df = pd.DataFrame(c, columns=trait_index, index=trait_index)
    return ResultTable(df, name=name)


def tab_credible_diff(diff,
                      diff_name,
                      x_labels,
                      cred_header='Cred.',
                      x_label='Population',
                      and_head=' ',
                      and_label='AND'):
    """Create table with credible trait differences between populations
    represented by included group data sets
    :param diff: list of tuples ((i, j), p), indicating that
        x_labels[i] > x-labels[j] with joint credibility p, meaning
        prob{ x_labels[i] > s_labels[j] } AND all previous pairs } == p
        in quantity = diff_name
    :param diff_name: string name identifying the difference shown
    :param x_labels: list of category labels of compared data
    :param x_label: string label of category for the difference
    :param cred_header: string header of credibility column
    :param and_head: name of column with AND labels
    :param and_label: label indicating AND condition
    :return: a ResultTable instance with header lines + one line for each credible difference,
    """
    if len(diff) == 0:
        return None
    col = {and_head: [''] + (len(diff) -1) * [and_label]}
    col |= {x_label + ' >': [x_labels[i]
                             for ((i, j), p) in diff]}  # category with larger value
    col |= {x_label: [x_labels[j]
                      for ((i, j), p) in diff]}  # category with smaller value
    col |= {cred_header: [p for ((i, j), p) in diff]}  # credibility value in (0, 1)
    df = pd.DataFrame(col)
    # df.items.name = diff_name  # suppressed anyway, by DiffTable.save()
    return DiffTable(df, name=diff_name)


# ------------------------------------------ internal help functions:
def _nice_ticks(y_min, y_max):
    """Make nice sequence of equally spaced ticks
    with at most one decimal
    :param y_min: scalar axis min
    :param y_max: scalar axis max
    :return: t = array with tick values
        all(y_min <= t <= y_max
    """
    d = y_max - y_min
    if d < 1.:
        step = 0.1
    elif d < 2.:
        step = 0.2
    elif d < 5.:
        step = 0.5
    else:
        step = 1.
    return np.arange(np.ceil(y_min / step), np.floor(y_max / step) + 1) * step


def _x_tick_style(labels):
    """Select xtick properties to avoid tick-label clutter
    :param labels: list of tick label strings
    :return: dict with keyword arguments for set_xticklabels
    """
    maxL = max(len(l) for l in labels)
    rotate_x_label = maxL * len(labels) > 60
    if rotate_x_label:
         style = dict(rotation=25, horizontalalignment='right')
    else:
        style = dict(rotation='horizontal', horizontalalignment='center')
    return style


def make_trait_labels(item_trait_map):  # ******** -> ir_model ?
    """Prepare trait labels for plots and tables
    :param item_trait_map: 2D boolean array mapping trait for each item
    :return: list of strings
    """
    # *** should be done externally, by ir_model.OrdinalResponseModel ? ***
    trait_sel = np.any(item_trait_map, axis=0)
    item_trait_map = item_trait_map[:, trait_sel]
    n_items = len(item_trait_map)
    t_items = [np.arange(1, n_items + 1, dtype=int)[t_i]
               for t_i in item_trait_map.T]
    return ['Q(' + ','.join(f'{i}' for i in t_i) + ')'
            for t_i in t_items]


def _clean_file_name(s):
    """Make a string that can be used as file name
    :param s: string
    :return: clean_s,
        with whitespace replaced by _,
        and not-allowed chars removed
    """
    clean_s = s.replace(' ', '_')
    return clean_s.translate({ord(c): None for c in '(),.*'})


# ------------------------------------------ help functions:
def safe_file_path(p):
    """Ensure previously non-existing file path, to avoid over-writing,
    by adding a sequence number to the path stem
    :param p: file path
    :return: p_safe = similar file path with modified name
    """
    f_stem = p.stem
    f_suffix = p.suffix
    i = 0
    while p.exists():
        i += 1
        p = p.with_name(f_stem + f'-{i}' + f_suffix)
    return p


# -----------------------------------------------------
if __name__ == '__main__':
    print(_clean_file_name('asd asdf. (asdf)*, asdf'))